import pathlib
import re
from io import StringIO

import numpy as np
import pytest
import zarr
from bio2zarr import vcf2zarr
from cyvcf2 import VCF
from numpy.testing import assert_array_equal

from vcztools.constants import INT_FILL, INT_MISSING
from vcztools.vcf_writer import _compute_info_fields, write_vcf

from .utils import assert_vcfs_close, vcz_path_cache


@pytest.mark.parametrize("output_is_path", [True, False])
def test_write_vcf(tmp_path, output_is_path):
    original = pathlib.Path("tests/data/vcf") / "sample.vcf.gz"
    vcz = vcz_path_cache(original)
    output = tmp_path.joinpath("output.vcf")

    if output_is_path:
        write_vcf(vcz, output, no_version=True)
    else:
        output_str = StringIO()
        write_vcf(vcz, output_str, no_version=True)
        with open(output, "w") as f:
            f.write(output_str.getvalue())

    v = VCF(output)

    assert v.samples == ["NA00001", "NA00002", "NA00003"]

    variant = next(v)

    assert variant.CHROM == "19"
    assert variant.POS == 111
    assert variant.ID is None
    assert variant.REF == "A"
    assert variant.ALT == ["C"]
    assert variant.QUAL == pytest.approx(9.6)
    assert variant.FILTER is None

    assert variant.genotypes == [[0, 0, True], [0, 0, True], [0, 1, False]]

    assert_array_equal(
        variant.format("HQ"),
        [[10, 15], [10, 10], [3, 3]],
    )

    # check headers are the same
    assert_vcfs_close(original, output)


@pytest.mark.parametrize(
    ("include", "exclude", "expected_chrom_pos"),
    [
        ("POS < 1000", None, [("19", 111), ("19", 112), ("X", 10)]),
        (
            None,
            "POS < 1000",
            [
                ("20", 14370),
                ("20", 17330),
                ("20", 1110696),
                ("20", 1230237),
                ("20", 1234567),
                ("20", 1235237),
            ],
        ),
    ],
)
def test_write_vcf__filtering(tmp_path, include, exclude, expected_chrom_pos):
    original = pathlib.Path("tests/data/vcf") / "sample.vcf.gz"
    vcz = vcz_path_cache(original)
    output = tmp_path.joinpath("output.vcf")

    write_vcf(vcz, output, include=include, exclude=exclude)

    v = VCF(str(output))
    variants = list(v)

    assert len(variants) == len(expected_chrom_pos)
    assert v.samples == ["NA00001", "NA00002", "NA00003"]

    for variant, chrom_pos in zip(variants, expected_chrom_pos):
        chrom, pos = chrom_pos
        assert variant.CHROM == chrom
        assert variant.POS == pos


# fmt: off
@pytest.mark.parametrize(
    ("regions", "targets", "expected_chrom_pos"),
    [
        # regions only
        ("19", None, [("19", 111), ("19", 112)]),
        ("19:112", None, [("19", 112)]),
        ("20:1230236-", None, [("20", 1230237), ("20", 1234567), ("20", 1235237)]),
        ("20:1230237-", None, [("20", 1230237), ("20", 1234567), ("20", 1235237)]),
        ("20:1230238-", None, [("20", 1234567), ("20", 1235237)]),
        ("20:1230237-1235236", None, [("20", 1230237), ("20", 1234567)]),
        ("20:1230237-1235237", None, [("20", 1230237), ("20", 1234567), ("20", 1235237)]),  # noqa: E501
        ("20:1230237-1235238", None, [("20", 1230237), ("20", 1234567), ("20", 1235237)]),  # noqa: E501
        ("19,X", None, [("19", 111), ("19", 112), ("X", 10)]),
        ("X:11", None, [("X", 10)]),  # note differs from targets

        # targets only
        (None, "19", [("19", 111), ("19", 112)]),
        (None, "19:112", [("19", 112)]),
        (None, "20:1230236-", [("20", 1230237), ("20", 1234567), ("20", 1235237)]),
        (None, "20:1230237-", [("20", 1230237), ("20", 1234567), ("20", 1235237)]),
        (None, "20:1230238-", [("20", 1234567), ("20", 1235237)]),
        (None, "20:1230237-1235236", [("20", 1230237), ("20", 1234567)]),
        (None, "20:1230237-1235237", [("20", 1230237), ("20", 1234567), ("20", 1235237)]),  # noqa: E501
        (None, "20:1230237-1235238", [("20", 1230237), ("20", 1234567), ("20", 1235237)]),  # noqa: E501
        (None, "19,X", [("19", 111), ("19", 112), ("X", 10)]),
        (None, "X:11", []),
        (None, "^19,20:1-1234567", [("20", 1235237), ("X", 10)]),  # complement

        # regions and targets
        ("20", "^20:1110696-", [("20", 14370), ("20", 17330)])
    ]
)
# fmt: on
def test_write_vcf__regions(tmp_path, regions, targets,
                            expected_chrom_pos):

    original = pathlib.Path("tests/data/vcf") / "sample.vcf.gz"
    vcz = vcz_path_cache(original)
    output = tmp_path.joinpath("output.vcf")

    write_vcf(vcz, output, variant_regions=regions,
              variant_targets=targets)

    v = VCF(output)
    variants = list(v)
    assert len(variants) == len(expected_chrom_pos)

    assert v.samples == ["NA00001", "NA00002", "NA00003"]

    for variant, chrom_pos in zip(variants, expected_chrom_pos):
        chrom, pos = chrom_pos
        assert variant.CHROM == chrom
        assert variant.POS == pos

@pytest.mark.parametrize(
    ("samples", "expected_samples", "expected_genotypes"),
    [
        ("NA00001", ["NA00001"], [[0, 0, True]]),
        ("NA00001,NA00003", ["NA00001", "NA00003"], [[0, 0, True], [0, 1, False]]),
        ("NA00003,NA00001", ["NA00003", "NA00001"], [[0, 1, False], [0, 0, True]]),
        ("^NA00002", ["NA00001", "NA00003"], [[0, 0, True], [0, 1, False]]),
        ("^NA00003,NA00002", ["NA00001"], [[0, 0, True]]),
        ("^NA00003,NA00002,NA00003", ["NA00001"], [[0, 0, True]]),
    ]
)
def test_write_vcf__samples(tmp_path, samples, expected_samples, expected_genotypes):
    original = pathlib.Path("tests/data/vcf") / "sample.vcf.gz"
    vcz = vcz_path_cache(original)
    output = tmp_path.joinpath("output.vcf")

    write_vcf(vcz, output, samples=samples)

    v = VCF(output)

    assert v.samples == expected_samples

    variant = next(v)

    assert variant.CHROM == "19"
    assert variant.POS == 111
    assert variant.ID is None
    assert variant.REF == "A"
    assert variant.ALT == ["C"]
    assert variant.QUAL == pytest.approx(9.6)
    assert variant.FILTER is None

    assert variant.genotypes == expected_genotypes


def test_write_vcf__no_samples(tmp_path):
    original = pathlib.Path("tests/data/vcf") / "sample.vcf.gz"
    vcz = vcz_path_cache(original)
    output = tmp_path.joinpath("output.vcf")

    write_vcf(vcz, output, drop_genotypes=True)

    v = VCF(output)

    assert v.samples == []


@pytest.mark.parametrize(
    ("regions", "targets", "samples", "include", "expected_chrom_pos"),
    [
        # Test that sample filtering takes place after include filtering.
        ("20", None, "NA00001", "FMT/GQ > 60", [("20", 1230237)]),
        # Test that region filtering and include expression are combined.
        ("19", None, "NA00001", "FMT/GQ > 60", []),
        # Test that target filtering and include expression are combined.
        (None, "19", "NA00001", "FMT/GQ > 60", [])
    ],
)
def test_write_vcf__regions_samples_filtering(
        tmp_path,
        regions,
        targets,
        samples,
        include,
        expected_chrom_pos
):
    original = pathlib.Path("tests/data/vcf") / "sample.vcf.gz"
    vcz = vcz_path_cache(original)
    output = tmp_path.joinpath("output.vcf")

    write_vcf(
        vcz,
        output,
        variant_regions=regions,
        variant_targets=targets,
        samples=samples,
        include=include)

    v = VCF(str(output))
    variants = list(v)

    assert len(variants) == len(expected_chrom_pos)
    assert v.samples == ["NA00001"]

    for variant, chrom_pos in zip(variants, expected_chrom_pos):
        chrom, pos = chrom_pos
        assert variant.CHROM == chrom
        assert variant.POS == pos


def test_write_vcf__include_exclude(tmp_path):
    original = pathlib.Path("tests/data/vcf") / "sample.vcf.gz"
    vcz = vcz_path_cache(original)
    output = tmp_path.joinpath("output.vcf")

    variant_site_filter = "POS > 1"

    with pytest.raises(
            ValueError,
            match=re.escape(
                "Cannot handle both an include expression and an exclude expression."
            )
    ):
        write_vcf(vcz, output, include=variant_site_filter, exclude=variant_site_filter)


def test_write_vcf__header_flags(tmp_path):
    original = pathlib.Path("tests/data/vcf") / "sample.vcf.gz"
    vcz = vcz_path_cache(original)
    output = tmp_path.joinpath("output.vcf")

    output_header = StringIO()
    write_vcf(vcz, output_header, header_only=True, no_version=True)

    output_no_header = StringIO()
    write_vcf(vcz, output_no_header, no_header=True, no_version=True)
    assert not output_no_header.getvalue().startswith("#")

    # combine outputs and check VCFs match
    output_str = output_header.getvalue() + output_no_header.getvalue()
    with open(output, "w") as f:
        f.write(output_str)
    assert_vcfs_close(original, output)


def test_write_vcf__generate_header(tmp_path):
    original = pathlib.Path("tests/data/vcf") / "sample.vcf.gz"
    # don't use cache here since we mutate the vcz
    vcz = tmp_path.joinpath("intermediate.vcz")
    vcf2zarr.convert([original], vcz, worker_processes=0, local_alleles=False)

    # remove vcf_header
    root = zarr.open(vcz, mode="r+")
    del root.attrs["vcf_header"]

    output_header = StringIO()
    write_vcf(vcz, output_header, header_only=True, no_version=True)

    expected_vcf_header = """##fileformat=VCFv4.3
##source={}
##INFO=<ID=AA,Number=1,Type=String,Description="Ancestral Allele">
##INFO=<ID=AC,Number=2,Type=Integer,Description="Allele count in genotypes, for each ALT allele, in the same order as listed">
##INFO=<ID=AF,Number=2,Type=Float,Description="Allele Frequency">
##INFO=<ID=AN,Number=1,Type=Integer,Description="Total number of alleles in called genotypes">
##INFO=<ID=DB,Number=0,Type=Flag,Description="dbSNP membership, build 129">
##INFO=<ID=DP,Number=1,Type=Integer,Description="Total Depth">
##INFO=<ID=H2,Number=0,Type=Flag,Description="HapMap2 membership">
##INFO=<ID=NS,Number=1,Type=Integer,Description="Number of Samples With Data">
##FILTER=<ID=PASS,Description="">
##FILTER=<ID=s50,Description="">
##FILTER=<ID=q10,Description="">
##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Read Depth">
##FORMAT=<ID=GQ,Number=1,Type=Integer,Description="Genotype Quality">
##FORMAT=<ID=HQ,Number=2,Type=Integer,Description="Haplotype Quality">
##contig=<ID=19>
##contig=<ID=20>
##contig=<ID=X>
#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO	FORMAT	NA00001	NA00002	NA00003
"""  # noqa: E501

    # substitute value of source
    expected_vcf_header = expected_vcf_header.format(root.attrs["source"])

    assert output_header.getvalue() == expected_vcf_header


def test_compute_info_fields():
    gt = np.array([
        [[0, 0], [0, 1], [1, 1]],
        [[0, 0], [0, 2], [2, 2]],
        [[0, 1], [1, 2], [2, 2]],
        [[INT_MISSING, INT_MISSING], [INT_MISSING, INT_MISSING], [INT_FILL, INT_FILL]],
        [[INT_MISSING, INT_MISSING], [0, 3], [INT_FILL, INT_FILL]],
    ])
    alt = np.array([
        [b"A", b"B", b""],
        [b"A", b"B", b"C"],
        [b"A", b"B", b"C"],
        [b"", b"", b""],
        [b"A", b"B", b"C"]
    ])
    expected_result = {
        "AC": np.array([
            [3, 0, INT_FILL],
            [0, 3, 0],
            [2, 3, 0],
            [INT_FILL, INT_FILL, INT_FILL],
            [0, 0, 1],
        ]),
        "AN": np.array([6, 6, 6, 0, 2]),
    }

    computed_info_fields = _compute_info_fields(gt, alt)

    assert expected_result.keys() == computed_info_fields.keys()

    for key in expected_result.keys():
        np.testing.assert_array_equal(expected_result[key], computed_info_fields[key])



import numpy as np
import zarr

from vcztools.utils import open_file_like


def nrecords(vcz, output):
    root = zarr.open(vcz, mode="r")

    with open_file_like(output) as output:
        num_variants = root["variant_position"].shape[0]
        print(num_variants, file=output)


def stats(vcz, output):
    root = zarr.open(vcz, mode="r")

    if "region_index" not in root:
        raise ValueError(
            "Could not load 'region_index' variable. "
            "Use 'vcz2zarr' to create an index."
        )

    with open_file_like(output) as output:
        contigs = root["contig_id"][:].astype("U").tolist()
        if "contig_length" in root:
            contig_lengths = root["contig_length"][:]
        else:
            contig_lengths = ["."] * len(contigs)

        region_index = root["region_index"][:]

        contig_indexes = region_index[:, 1]
        num_records = region_index[:, 5]

        num_records_per_contig = np.bincount(
            contig_indexes, weights=num_records
        ).astype(np.int64)

        for contig, contig_length, nr in zip(
            contigs, contig_lengths, num_records_per_contig
        ):
            if nr > 0:
                print(f"{contig}\t{contig_length}\t{nr}", file=output)



import pathlib
from collections.abc import Iterator
from contextlib import contextmanager
from itertools import zip_longest

import cyvcf2
import numpy as np
from bio2zarr import vcf2zarr


@contextmanager
def open_vcf(path) -> Iterator[cyvcf2.VCF]:
    """A context manager for opening a VCF file."""
    vcf = cyvcf2.VCF(path)
    try:
        yield vcf
    finally:
        vcf.close()


def normalise_info_missingness(info_dict, key):
    value = info_dict.get(key, None)
    if isinstance(value, tuple):
        if all(x is None for x in value):
            value = None
    elif isinstance(value, str):
        if all(x == "." for x in value.split(",")):
            value = None
    return value


def assert_vcfs_close(f1, f2, *, rtol=1e-05, atol=1e-03):
    """Like :py:func:`numpy.testing.assert_allclose()`, but for VCF files.

    Raises an `AssertionError` if two VCF files are not equal to one another.
    Float values in QUAL, INFO, or FORMAT fields are compared up to the
    desired tolerance. All other values must match exactly.

    Parameters
    ----------
    f1
        Path to first VCF to compare.
    f2
        Path to second VCF to compare.
    rtol
        Relative tolerance.
    atol
        Absolute tolerance.
    """
    with open_vcf(f1) as vcf1, open_vcf(f2) as vcf2:
        assert vcf1.raw_header == vcf2.raw_header
        assert vcf1.samples == vcf2.samples

        for v1, v2 in zip_longest(vcf1, vcf2):
            if v1 is None and v2 is not None:
                raise AssertionError(f"Right contains extra variant: {v2}")
            if v1 is not None and v2 is None:
                raise AssertionError(f"Left contains extra variant: {v1}")

            assert v1.CHROM == v2.CHROM, f"CHROM not equal for variants\n{v1}{v2}"
            assert v1.POS == v2.POS, f"POS not equal for variants\n{v1}{v2}"
            assert v1.ID == v2.ID, f"ID not equal for variants\n{v1}{v2}"
            assert v1.REF == v2.REF, f"REF not equal for variants\n{v1}{v2}"
            assert v1.ALT == v2.ALT, f"ALT not equal for variants\n{v1}{v2}"
            np.testing.assert_allclose(
                np.array(v1.QUAL, dtype=np.float32),
                np.array(v2.QUAL, dtype=np.float32),
                rtol=rtol,
                atol=atol,
                err_msg=f"QUAL not equal for variants\n{v1}{v2}",
            )
            assert set(v1.FILTERS) == set(
                v2.FILTERS
            ), f"FILTER not equal for variants\n{v1}{v2}"

            v1_info = dict(v1.INFO)
            v2_info = dict(v2.INFO)
            all_keys = set(v1_info.keys()) | set(v2_info.keys())
            for k in all_keys:
                val1 = normalise_info_missingness(v1_info, k)
                val2 = normalise_info_missingness(v2_info, k)
                # values are python objects (not np arrays)
                if isinstance(val1, float) or (
                    isinstance(val1, tuple) and any(isinstance(v, float) for v in val1)
                ):
                    np.testing.assert_allclose(
                        np.array(val1, dtype=np.float32),
                        np.array(val2, dtype=np.float32),
                        rtol=rtol,
                        atol=atol,
                        err_msg=f"INFO {k} not equal for variants\n{v1}{v2}",
                    )
                else:
                    assert val1 == val2, f"INFO {k} not equal for variants\n{v1}{v2}"

            # NOTE skipping this because it requires items to be in the same order.
            # assert v1.FORMAT == v2.FORMAT, f"FORMAT not equal for variants\n{v1}{v2}"
            for field in v1.FORMAT:
                if field == "GT":
                    assert (
                        v1.genotypes == v2.genotypes
                    ), f"GT not equal for variants\n{v1}{v2}"
                else:
                    val1 = v1.format(field)
                    val2 = v2.format(field)
                    if val2 is None:
                        # FIXME this is a quick hack to workaround missing support for
                        # dealing with the field missing vs all-elements-in-field
                        # missing issue.
                        # https://github.com/jeromekelleher/vcztools/issues/14
                        assert [str(x) == "." for x in val1]
                    else:
                        if val1.dtype.kind == "f":
                            np.testing.assert_allclose(
                                val1,
                                val2,
                                rtol=rtol,
                                atol=atol,
                                err_msg=f"FORMAT {field} not equal for "
                                f"variants\n{v1}{v2}",
                            )
                        else:
                            np.testing.assert_array_equal(
                                val1,
                                val2,
                                err_msg=f"FORMAT {field} not equal for "
                                f"variants\n{v1}{v2}",
                            )


def vcz_path_cache(vcf_path):
    """
    Store converted files in a cache to speed up tests. We're not testing
    vcf2zarr here, so no point in running over and over again.
    """
    cache_path = pathlib.Path("vcz_test_cache")
    if not cache_path.exists():
        cache_path.mkdir()
    cached_vcz_path = (cache_path / vcf_path.name).with_suffix(".vcz")
    if not cached_vcz_path.exists():
        if vcf_path.name.startswith("chr22"):
            vcf2zarr.convert(
                [vcf_path],
                cached_vcz_path,
                worker_processes=0,
                variants_chunk_size=10,
                samples_chunk_size=10,
            )
        else:
            vcf2zarr.convert(
                [vcf_path], cached_vcz_path, worker_processes=0, local_alleles=False
            )
    return cached_vcz_path



import re
from typing import Any, Optional

import numpy as np
import pandas as pd
from pyranges import PyRanges


def parse_region_string(region: str) -> tuple[str, Optional[int], Optional[int]]:
    """Return the contig, start position and end position from a region string."""
    if re.search(r":\d+-\d*$", region):
        contig, start_end = region.rsplit(":", 1)
        start, end = start_end.split("-")
        return contig, int(start), int(end) if len(end) > 0 else None
    elif re.search(r":\d+$", region):
        contig, start = region.rsplit(":", 1)
        return contig, int(start), int(start)
    else:
        contig = region
        return contig, None, None


def regions_to_pyranges(
    regions: list[tuple[str, Optional[int], Optional[int]]], all_contigs: list[str]
) -> PyRanges:
    """Convert region tuples to a PyRanges object."""

    chromosomes = []
    starts = []
    ends = []
    for contig, start, end in regions:
        if start is None:
            start = 0
        else:
            start -= 1

        if end is None:
            end = np.iinfo(np.int64).max

        chromosomes.append(all_contigs.index(contig))
        starts.append(start)
        ends.append(end)

    return PyRanges(chromosomes=chromosomes, starts=starts, ends=ends)


def parse_regions(regions: Optional[str], all_contigs: list[str]) -> Optional[PyRanges]:
    """Return a PyRanges object from a comma-separated set of region strings."""
    if regions is None:
        return None
    return regions_to_pyranges(
        [parse_region_string(region) for region in regions.split(",")], all_contigs
    )


def parse_targets(
    targets: Optional[str], all_contigs: list[str]
) -> tuple[Optional[PyRanges], bool]:
    """Return a PyRanges object from a comma-separated set of region strings,
    optionally preceeded by a ^ character to indicate complement."""
    if targets is None:
        return None, False
    complement = targets.startswith("^")
    return (
        parse_regions(targets[1:] if complement else targets, all_contigs),
        complement,
    )


def regions_to_chunk_indexes(
    regions: Optional[PyRanges],
    targets: Optional[PyRanges],
    complement: bool,
    regions_index: Any,
):
    """Return chunks indexes that overlap the given regions or targets.

    If both regions and targets are specified then only regions are used
    to find overlapping chunks (since targets are used later to refine).

    If only targets are specified then they are used to find overlapping chunks,
    taking into account the complement flag.
    """

    # Create pyranges for chunks using the region index.
    # For regions use max end position, for targets just end position
    chunk_index = regions_index[:, 0]
    contig_id = regions_index[:, 1]
    start_position = regions_index[:, 2]
    end_position = regions_index[:, 3]
    max_end_position = regions_index[:, 4]
    df = pd.DataFrame(
        {
            "chunk_index": chunk_index,
            "Chromosome": contig_id,
            "Start": start_position,
            "End": max_end_position if regions is not None else end_position,
        }
    )
    chunk_regions = PyRanges(df)

    if regions is not None:
        overlap = chunk_regions.overlap(regions)
    elif complement:
        overlap = chunk_regions.subtract(targets)
    else:
        overlap = chunk_regions.overlap(targets)
    if overlap.empty:
        return np.empty((0,), dtype=np.int64)
    chunk_indexes = overlap.df["chunk_index"].to_numpy()
    chunk_indexes = np.unique(chunk_indexes)
    return chunk_indexes


def regions_to_selection(
    regions: Optional[PyRanges],
    targets: Optional[PyRanges],
    complement: bool,
    variant_contig: Any,
    variant_position: Any,
    variant_length: Any,
):
    """Return a variant selection that corresponds to the given regions and targets.

    If both regions and targets are specified then they are both used to find
    overlapping variants.
    """

    # subtract 1 from start coordinate to convert intervals
    # from VCF (1-based, fully-closed) to Python (0-based, half-open)
    variant_start = variant_position - 1

    if regions is not None:
        variant_end = variant_start + variant_length
        df = pd.DataFrame(
            {"Chromosome": variant_contig, "Start": variant_start, "End": variant_end}
        )
        # save original index as column so we can retrieve it after finding overlap
        df["index"] = df.index
        variant_regions = PyRanges(df)
    else:
        variant_regions = None

    if targets is not None:
        targets_variant_end = variant_position  # length 1
        df = pd.DataFrame(
            {
                "Chromosome": variant_contig,
                "Start": variant_start,
                "End": targets_variant_end,
            }
        )
        # save original index as column so we can retrieve it after finding overlap
        df["index"] = df.index
        variant_targets = PyRanges(df)
    else:
        variant_targets = None

    if variant_regions is not None:
        regions_overlap = variant_regions.overlap(regions)
    else:
        regions_overlap = None

    if variant_targets is not None:
        if complement:
            targets_overlap = variant_targets.subtract(targets)
        else:
            targets_overlap = variant_targets.overlap(targets)
    else:
        targets_overlap = None

    if regions_overlap is not None and targets_overlap is not None:
        overlap = regions_overlap.overlap(targets_overlap)
    elif regions_overlap is not None:
        overlap = regions_overlap
    else:
        overlap = targets_overlap

    if overlap.empty:
        return np.empty((0,), dtype=np.int64)
    return overlap.df["index"].to_numpy()



import numpy as np

INT_MISSING, INT_FILL = -1, -2

FLOAT32_MISSING, FLOAT32_FILL = np.array([0x7F800001, 0x7F800002], dtype=np.int32).view(
    np.float32
)
FLOAT32_MISSING_AS_INT32, FLOAT32_FILL_AS_INT32 = np.array(
    [0x7F800001, 0x7F800002], dtype=np.int32
)

# From VCF fixed fields
RESERVED_VARIABLE_NAMES = [
    "variant_contig",
    "variant_position",
    "variant_length",
    "variant_id",
    "variant_id_mask",
    "variant_allele",
    "variant_quality",
    "variant_filter",
]

RESERVED_VCF_FIELDS = {
    "CHROM": "variant_contig",
    "POS": "variant_position",
    "ID": "variant_id",
    "REF": "variant_allele",
    "ALT": "variant_allele",
    "QUAL": "variant_quality",
    "FILTER": "variant_filter",
}



import pytest
from numpy.testing import assert_array_equal

from vcztools.utils import (
    search,
    vcf_name_to_vcz_name,
)


@pytest.mark.parametrize(
    ("a", "v", "expected_ind"),
    [
        (["a", "b", "c", "d"], ["b", "a", "c"], [1, 0, 2]),
        (["a", "c", "d", "b"], ["b", "a", "c"], [3, 0, 1]),
        (["a", "c", "d", "b"], ["b", "a", "a", "c"], [3, 0, 0, 1]),
        (["a", "c", "d", "b"], [], []),
    ],
)
def test_search(a, v, expected_ind):
    assert_array_equal(search(a, v), expected_ind)


@pytest.mark.parametrize(
    ("vczs", "vcf", "expected_vcz"),
    [
        (set(), "GT", "call_genotype"),
        (set(), "FMT/GT", "call_genotype"),
        ({"call_DP"}, "DP", "call_DP"),
        ({"variant_DP"}, "DP", "variant_DP"),
        ({"call_DP", "variant_DP"}, "DP", "call_DP"),
        ({"call_DP", "variant_DP"}, "INFO/DP", "variant_DP"),
        ({"call_DP", "variant_DP"}, "FORMAT/DP", "call_DP"),
        ({"variant_DP"}, "FORMAT/DP", "call_DP"),
        (set(), "CHROM", "variant_contig"),
        (set(), "POS", "variant_position"),
        (set(), "ID", "variant_id"),
        (set(), "REF", "variant_allele"),
        (set(), "ALT", "variant_allele"),
        (set(), "QUAL", "variant_quality"),
        (set(), "FILTER", "variant_filter"),
    ],
)
def test_vcf_to_vcz(vczs, vcf, expected_vcz):
    assert vcf_name_to_vcz_name(vczs, vcf) == expected_vcz



import pathlib

import numpy as np
import pyparsing as pp
import pytest
import zarr
from numpy.testing import assert_array_equal

from tests.utils import vcz_path_cache
from vcztools.filter import FilterExpressionEvaluator, FilterExpressionParser


class TestFilterExpressionParser:
    @pytest.fixture()
    def identifier_parser(self, parser):
        return parser._identifier_parser

    @pytest.fixture()
    def parser(self):
        return FilterExpressionParser()

    @pytest.mark.parametrize(
        ("expression", "expected_result"),
        [
            ("1", [1]),
            ("1.0", [1.0]),
            ("1e-4", [0.0001]),
            ('"String"', ["String"]),
            ("POS", ["POS"]),
            ("INFO/DP", ["INFO/DP"]),
            ("FORMAT/GT", ["FORMAT/GT"]),
            ("FMT/GT", ["FMT/GT"]),
            ("GT", ["GT"]),
        ],
    )
    def test_valid_identifiers(self, identifier_parser, expression, expected_result):
        assert identifier_parser(expression).as_list() == expected_result

    @pytest.mark.parametrize(
        "expression",
        [
            "",
            "FORMAT/ GT",
            "format / GT",
            "fmt / GT",
            "info / DP",
            "'String'",
        ],
    )
    def test_invalid_identifiers(self, identifier_parser, expression):
        with pytest.raises(pp.ParseException):
            identifier_parser(expression)

    @pytest.mark.parametrize(
        ("expression", "expected_result"),
        [
            ("POS>=100", [["POS", ">=", 100]]),
            (
                "FMT/DP>10 && FMT/GQ>10",
                [[["FMT/DP", ">", 10], "&&", ["FMT/GQ", ">", 10]]],
            ),
            ("QUAL>10 || FMT/GQ>10", [[["QUAL", ">", 10], "||", ["FMT/GQ", ">", 10]]]),
            (
                "FMT/DP>10 && FMT/GQ>10 || QUAL > 10",
                [
                    [
                        [["FMT/DP", ">", 10], "&&", ["FMT/GQ", ">", 10]],
                        "||",
                        ["QUAL", ">", 10],
                    ]
                ],
            ),
            (
                "QUAL>10 || FMT/DP>10 && FMT/GQ>10",
                [
                    [
                        ["QUAL", ">", 10],
                        "||",
                        [["FMT/DP", ">", 10], "&&", ["FMT/GQ", ">", 10]],
                    ]
                ],
            ),
            (
                "QUAL>10 | FMT/DP>10 & FMT/GQ>10",
                [
                    [
                        ["QUAL", ">", 10],
                        "|",
                        [["FMT/DP", ">", 10], "&", ["FMT/GQ", ">", 10]],
                    ],
                ],
            ),
            (
                "(QUAL>10 || FMT/DP>10) && FMT/GQ>10",
                [
                    [
                        [["QUAL", ">", 10], "||", ["FMT/DP", ">", 10]],
                        "&&",
                        ["FMT/GQ", ">", 10],
                    ]
                ],
            ),
        ],
    )
    def test_valid_expressions(self, parser, expression, expected_result):
        assert parser(expression=expression).as_list() == expected_result


class TestFilterExpressionEvaluator:
    @pytest.mark.parametrize(
        ("expression", "expected_result"),
        [
            ("POS < 1000", [1, 1, 0, 0, 0, 0, 0, 0, 1]),
            ("FMT/GQ > 20", [0, 0, 1, 1, 1, 1, 1, 0, 0]),
            ("FMT/DP >= 5 && FMT/GQ > 10", [0, 0, 1, 1, 1, 0, 0, 0, 0]),
            ("FMT/DP >= 5 & FMT/GQ>10", [0, 0, 1, 0, 1, 0, 0, 0, 0]),
            ("QUAL > 10 || FMT/GQ>10", [0, 0, 1, 1, 1, 1, 1, 0, 0]),
            ("(QUAL > 10 || FMT/GQ>10) && POS > 100000", [0, 0, 0, 0, 1, 1, 1, 0, 0]),
            ("(FMT/DP >= 8 | FMT/GQ>40) && POS > 100000", [0, 0, 0, 0, 0, 1, 0, 0, 0]),
            ("INFO/DP > 10", [0, 0, 1, 1, 0, 1, 0, 0, 0]),
            ("GT > 0", [1, 1, 1, 1, 1, 0, 1, 0, 1]),
            ("GT > 0 & FMT/HQ >= 10", [0, 0, 1, 1, 1, 0, 0, 0, 0]),
        ],
    )
    def test(self, expression, expected_result):
        original = pathlib.Path("tests/data/vcf") / "sample.vcf.gz"
        vcz = vcz_path_cache(original)
        root = zarr.open(vcz, mode="r")

        parser = FilterExpressionParser()
        parse_results = parser(expression)[0]
        evaluator = FilterExpressionEvaluator(parse_results)
        assert_array_equal(evaluator(root, 0), expected_result)

        invert_evaluator = FilterExpressionEvaluator(parse_results, invert=True)
        assert_array_equal(invert_evaluator(root, 0), np.logical_not(expected_result))



import sys

import _vcztools
import numpy as np
import zarr

# From VCF fixed fields
RESERVED_VARIABLE_NAMES = [
    "variant_contig",
    "variant_position",
    "variant_id",
    "variant_id_mask",
    "variant_allele",
    "variant_quality",
    "variant_filter",
]


def copy_to_memory(group):
    mem_group = zarr.group()
    for name, array in group.items():
        if array.dtype == "O":
            copy = mem_group.empty_like(name, array, compressor={})
            # FIXME: this is the best I've been able to come up with here.
            # Something very weird with object arrays in v2
            for j, row in enumerate(array):
                copy[j] = row
        else:
            copy = mem_group.empty_like(name, array, compressor={})
            copy[:] = array[:]
        # print("copy = ", copy[:])
    return mem_group


def main(root):
    v_chunk = 0
    contigs = root["contig_id"][:].astype("S")
    # filters = root["filter_id"][:].astype("S")
    # print("contigs = ", contigs)
    # print("filters = ", contigs)

    chrom = contigs[root.variant_contig.blocks[v_chunk]]
    pos = root.variant_position.blocks[v_chunk]
    id = root.variant_id.blocks[v_chunk].astype("S")
    alleles = root.variant_allele.blocks[v_chunk]
    ref = alleles[:, 0].astype("S")
    alt = alleles[:, 1:].astype("S")
    # qual = root.variant_quality.blocks[v_chunk]
    # filter_ = filters[root.variant_filter.blocks[v_chunk]]

    num_variants = len(pos)
    if len(id.shape) == 1:
        id = id.reshape((num_variants, 1))

    # TODO gathering fields and doing IO will be done separately later so that
    # we avoid retrieving stuff we don't need.
    format_fields = {}
    info_fields = {}
    for name, array in root.arrays():
        if name.startswith("call_") and not name.startswith("call_genotype"):
            vcf_name = name[len("call_") :]
            format_fields[vcf_name] = array.blocks[v_chunk]
        elif name.startswith("variant_") and name not in RESERVED_VARIABLE_NAMES:
            vcf_name = name[len("variant_") :]
            info_fields[vcf_name] = array.blocks[v_chunk]

    gt = None
    gt_phased = None
    if "call_genotype" in root:
        array = root["call_genotype"]
        gt = array.blocks[v_chunk]
        if "call_genotype_phased" in root:
            array = root["call_genotype_phased"]
            gt_phased = array.blocks[v_chunk]
        else:
            gt_phased = np.zeros_like(gt, dtype=bool)

    # print(gt, gt_phased)
    # print(list(format_fields.keys()))
    # print(list(info_fields.keys()))

    # print(contigs[chrom])
    # print(bytes(contigs[chrom]))
    # print(pos)
    # print(alleles)
    # print(alleles.dtype)
    # print(chrom)
    # print(pos)
    # print(id)
    # print(ref)
    # print(alt)

    num_samples = 0
    if gt is not None:
        num_samples = gt.shape[1]

    encoder = _vcztools.VcfEncoder(
        num_variants, num_samples, chrom=chrom, pos=pos, id=id, alt=alt, ref=ref
    )
    print(gt.shape)
    print(gt_phased.shape)
    encoder.add_gt_field(gt.astype("int32"), gt_phased)
    # # print(encoder.arrays)
    # # print(encoder)
    for name, array in info_fields.items():
        if array.dtype.kind == "O":
            array = array.astype("S")
        if len(array.shape) == 1:
            array = array.reshape((num_variants, 1))
        if array.dtype.kind == "i":
            array = array.astype("int32")  # tmp
        if array.dtype.kind == "f":
            continue  # tmp
        if array.dtype.kind == "b":
            continue  # tmp
            # array = array.astype("int32") # tmp

        print(name, array.dtype, array.dtype.kind)
        encoder.add_info_field(name, array)

    for name, array in format_fields.items():
        if array.dtype.kind == "O":
            array = array.astype("S")
        if len(array.shape) == 2:
            array = array.reshape((num_variants, num_samples, 1))
        if array.dtype.kind == "i":
            array = array.astype("int32")  # tmp
        if array.dtype.kind == "f":
            continue  # tmp
            # array = array.astype("int32") # tmp

        print(name, array.dtype, array.dtype.kind)
        encoder.add_format_field(name, array)

    # d = encoder.arrays
    # pos = encoder.arrays["POS"]
    # print(pos)
    # # print(d)
    # pos[0] = 123457
    # print(pos.flags)
    # pos.resize(0, refcheck=False)
    # print(pos)

    encoder.print_state(sys.stdout)
    for k, v in encoder.arrays.items():
        print(k, "\t", v.shape)
    for j in range(num_variants):
        line = encoder.encode_row(j, 2**30)
        print(line)


if __name__ == "__main__":
    root = zarr.open(sys.argv[1], mode="r")
    # root = copy_to_memory(root)
    # print("pos = ", root["variant_position"].info)
    # print(root.tree())
    main(root)
    # for _ in range(10000):
    # import tqdm

    # for _ in tqdm.tqdm(range(10000)):
    #     main(root)



from contextlib import ExitStack, contextmanager
from pathlib import Path

import numpy as np

from vcztools.constants import RESERVED_VCF_FIELDS


def search(a, v):
    """Like `np.searchsorted`, but array `a` does not have to be sorted."""
    sorter = np.argsort(a)
    rank = np.searchsorted(a, v, sorter=sorter)
    return sorter[rank]


@contextmanager
def open_file_like(file):
    """A context manager for opening a file path or string (and closing on exit),
    or passing a file-like object through."""
    with ExitStack() as stack:
        if isinstance(file, (str, Path)):
            file = stack.enter_context(open(file, mode="w"))
        yield file


def vcf_name_to_vcz_name(vcz_names: set[str], vcf_name: str) -> str:
    """
    Convert the name of a VCF field to the name of the corresponding VCF Zarr array.

    :param set[str] vcz_names: A set of potential VCF Zarr field names
    :param str vcf_name: The name of the VCF field
    :return: The name of the corresponding VCF Zarr array
    :rtype: str
    """
    split = vcf_name.split("/")
    assert 1 <= len(split) <= 2
    is_genotype_field = split[-1] == "GT"
    is_format_field = (
        split[0] in {"FORMAT", "FMT"}
        if len(split) > 1
        else is_genotype_field or f"call_{split[-1]}" in vcz_names
    )
    is_info_field = split[0] == "INFO" or split[-1] not in RESERVED_VCF_FIELDS

    if is_format_field:
        if is_genotype_field:
            return "call_genotype"
        else:
            return "call_" + split[-1]
    elif is_info_field:
        return f"variant_{split[-1]}"
    else:
        return RESERVED_VCF_FIELDS[vcf_name]



import contextlib
import os
import sys

import click

from . import query as query_module
from . import stats as stats_module
from . import vcf_writer


@contextlib.contextmanager
def handle_broken_pipe(output):
    """
    Handle sigpipe following official advice:
    https://docs.python.org/3/library/signal.html#note-on-sigpipe
    """
    try:
        yield
        # flush output here to force SIGPIPE to be triggered
        # while inside this try block.
        output.flush()
    except BrokenPipeError:
        # Python flushes standard streams on exit; redirect remaining output
        # to devnull to avoid another BrokenPipeError at shutdown
        devnull = os.open(os.devnull, os.O_WRONLY)
        os.dup2(devnull, sys.stdout.fileno())
        sys.exit(1)  # Python exits with error code 1 on EPIPE


include = click.option(
    "-i", "--include", type=str, help="Filter expression to include variant sites."
)
exclude = click.option(
    "-e", "--exclude", type=str, help="Filter expression to exclude variant sites."
)
output = click.option(
    "-o",
    "--output",
    type=click.File("w"),
    default="-",
    help="File path to write output to (defaults to stdout '-').",
)


class NaturalOrderGroup(click.Group):
    """
    List commands in the order they are provided in the help text.
    """

    def list_commands(self, ctx):
        return self.commands.keys()


@click.command
@click.argument("path", type=click.Path())
@click.option(
    "-n",
    "--nrecords",
    is_flag=True,
    help="Print the number of records (variants).",
)
@click.option(
    "-s",
    "--stats",
    is_flag=True,
    help="Print per contig stats.",
)
def index(path, nrecords, stats):
    if nrecords and stats:
        raise click.UsageError("Expected only one of --stats or --nrecords options")
    if nrecords:
        stats_module.nrecords(path, sys.stdout)
    elif stats:
        stats_module.stats(path, sys.stdout)
    else:
        raise click.UsageError("Building region indexes is not supported")


@click.command
@click.argument("path", type=click.Path())
@output
@click.option(
    "-l",
    "--list-samples",
    is_flag=True,
    help="List the sample IDs and exit.",
)
@click.option(
    "-f",
    "--format",
    type=str,
    help="The format of the output.",
    default=None,
)
@include
@exclude
def query(path, output, list_samples, format, include, exclude):
    if list_samples:
        # bcftools query -l ignores the --output option and always writes to stdout
        output = sys.stdout
        with handle_broken_pipe(output):
            query_module.list_samples(path, output)
        return

    if format is None:
        raise click.UsageError("Missing option -f / --format")
    with handle_broken_pipe(output):
        query_module.write_query(
            path, output, query_format=format, include=include, exclude=exclude
        )


@click.command
@click.argument("path", type=click.Path())
@output
@click.option(
    "-h",
    "--header-only",
    is_flag=True,
    help="Output the VCF header only.",
)
@click.option(
    "-H",
    "--no-header",
    is_flag=True,
    help="Suppress the header in VCF output.",
)
@click.option(
    "--no-version",
    is_flag=True,
    help="Do not append version and command line information to the output VCF header.",
)
@click.option(
    "-r",
    "--regions",
    type=str,
    default=None,
    help="Regions to include.",
)
# @click.option(
#     "--force-samples", is_flag=True, help="Only warn about unknown sample subsets."
# )
# @click.option(
#     "-I",
#     "--no-update",
#     is_flag=True,
#     help="Do not recalculate INFO fields for the sample subset.",
# )
# @click.option(
#     "-s",
#     "--samples",
#     type=str,
#     default=None,
#     help="Samples to include.",
# )
# @click.option(
#     "-S",
#     "--samples-file",
#     type=str,
#     default=None,
#     help="File of sample names to include.",
# )
@click.option(
    "-G",
    "--drop-genotypes",
    type=bool,
    is_flag=True,
    help="Drop genotypes.",
)
@click.option(
    "-t",
    "--targets",
    type=str,
    default=None,
    help="Target regions to include.",
)
@include
@exclude
def view(
    path,
    output,
    header_only,
    no_header,
    no_version,
    regions,
    targets,
    # force_samples,
    # no_update,
    # samples,
    # samples_file,
    drop_genotypes,
    include,
    exclude,
):
    suffix = output.name.split(".")[-1]
    # Exclude suffixes which require bgzipped or BCF output:
    # https://github.com/samtools/htslib/blob/329e7943b7ba3f0af15b0eaa00a367a1ac15bd83/vcf.c#L3815
    if suffix in ["gz", "bcf", "bgz"]:
        raise ValueError(
            f"Only uncompressed VCF output supported, suffix .{suffix} not allowed"
        )

    # Dropping implementation here until it's reimplemented after initial release:
    # https://github.com/sgkit-dev/vcztools/issues/121
    # if samples_file:
    #     assert not samples, "vcztools does not support combining -s and -S"

    #     samples = ""
    #     exclude_samples_file = samples_file.startswith("^")
    #     samples_file = samples_file.lstrip("^")

    #     with open(samples_file) as file:
    #         if exclude_samples_file:
    #             samples = "^" + samples
    #         samples += ",".join(line.strip() for line in file.readlines())

    with handle_broken_pipe(output):
        vcf_writer.write_vcf(
            path,
            output,
            header_only=header_only,
            no_header=no_header,
            no_version=no_version,
            variant_regions=regions,
            variant_targets=targets,
            # no_update=no_update,
            # samples=samples,
            drop_genotypes=drop_genotypes,
            include=include,
            exclude=exclude,
        )


@click.group(cls=NaturalOrderGroup, name="vcztools")
def vcztools_main():
    pass


vcztools_main.add_command(index)
vcztools_main.add_command(query)
vcztools_main.add_command(view)



import pathlib
from io import StringIO

import pytest
import zarr
from bio2zarr import vcf2zarr

from vcztools.stats import nrecords, stats

from .utils import vcz_path_cache


def test_nrecords():
    original = pathlib.Path("tests/data/vcf") / "sample.vcf.gz"
    vcz = vcz_path_cache(original)

    output_str = StringIO()
    nrecords(vcz, output_str)
    assert output_str.getvalue() == "9\n"


def test_stats():
    original = pathlib.Path("tests/data/vcf") / "sample.vcf.gz"
    vcz = vcz_path_cache(original)

    output_str = StringIO()
    stats(vcz, output_str)

    assert (
        output_str.getvalue()
        == """19	.	2
20	.	6
X	.	1
"""
    )


def test_stats__no_index(tmp_path):
    original = pathlib.Path("tests/data/vcf") / "sample.vcf.gz"
    # don't use cache here since we want to make sure vcz is not indexed
    vcz = tmp_path.joinpath("intermediate.vcz")
    vcf2zarr.convert([original], vcz, worker_processes=0, local_alleles=False)

    # delete the index created by vcf2zarr
    root = zarr.open(vcz, mode="a")
    del root["region_index"]

    with pytest.raises(ValueError, match="Could not load 'region_index' variable."):
        stats(vcz, StringIO())



# This script requires pv.

# https://www.ivarch.com/programs/pv.shtml

# There is a Homebrew formula to install pv on macOS.

# This script also depends on the simulation data:
# make -C data

import subprocess
import sys


def run_time_pv(command: str):
    print(command)
    subprocess.run(f"time {command} | pv > /dev/null", shell=True)
    print()


def run_bcftools(command: str, dataset_name: str):
    run_time_pv(f"bcftools {command} data/{dataset_name}.vcf.gz")


def run_vcztools(command: str, dataset_name: str):
    run_time_pv(f"vcztools {command} data/{dataset_name}.vcz")


if __name__ == "__main__":
    commands = [
        ("view", "sim_10k"),
        ("view", "chr22"),
        ("view -s tsk_7068,tsk_8769,tsk_8820", "sim_10k"),
        (r"query -f '%CHROM %POS %REF %ALT{0}\n'", "sim_10k"),
        (r"query -f '%CHROM:%POS\n' -i 'POS=49887394 | POS=50816415'", "sim_10k"),
        ("view -s '' --force-samples", "sim_10k"),
        ("view -i 'FMT/DP>10 & FMT/GQ>10'", "chr22"),
        ("view -i 'QUAL>10 || FMT/GQ>10'", "chr22"),
        (r"query -f 'GQ:[ %GQ] \t GT:[ %GT]\n'", "chr22"),
    ]

    if len(sys.argv) == 2 and sys.argv[1].isnumeric():
        index = int(sys.argv[1])
        command, dataset = commands[index]
        run_bcftools(command, dataset)
        run_vcztools(command, dataset)
    else:
        for command, dataset in commands:
            run_bcftools(command, dataset)
            run_vcztools(command, dataset)



import pathlib
import re
from io import StringIO

import pyparsing as pp
import pytest
import zarr

from tests.utils import vcz_path_cache
from vcztools.query import (
    QueryFormatGenerator,
    QueryFormatParser,
    list_samples,
    write_query,
)


def test_list_samples(tmp_path):
    vcf_path = pathlib.Path("tests/data/vcf") / "sample.vcf.gz"
    vcz_path = vcz_path_cache(vcf_path)
    expected_output = "NA00001\nNA00002\nNA00003\n"

    with StringIO() as output:
        list_samples(vcz_path, output)
        assert output.getvalue() == expected_output


class TestQueryFormatParser:
    @pytest.fixture()
    def parser(self):
        return QueryFormatParser()

    @pytest.mark.parametrize(
        ("expression", "expected_result"),
        [
            ("%CHROM", ["%CHROM"]),
            (r"\n", ["\n"]),
            (r"\t", ["\t"]),
            (r"%CHROM\n", ["%CHROM", "\n"]),
            ("%CHROM  %POS  %REF", ["%CHROM", "  ", "%POS", "  ", "%REF"]),
            (r"%CHROM  %POS0  %REF\n", ["%CHROM", "  ", "%POS0", "  ", "%REF", "\n"]),
            (
                r"%CHROM\t%POS\t%REF\t%ALT{0}\n",
                ["%CHROM", "\t", "%POS", "\t", "%REF", "\t", ["%ALT", 0], "\n"],
            ),
            (
                r"%CHROM\t%POS0\t%END\t%ID\n",
                ["%CHROM", "\t", "%POS0", "\t", "%END", "\t", "%ID", "\n"],
            ),
            (r"%CHROM:%POS\n", ["%CHROM", ":", "%POS", "\n"]),
            (r"%AC{1}\n", [["%AC", 1], "\n"]),
            (
                r"Read depth: %INFO/DP\n",
                ["Read", " ", "depth:", " ", "%INFO/DP", "\n"],
            ),
            (
                r"%CHROM\t%POS\t%REF\t%ALT[\t%SAMPLE=%GT]\n",
                [
                    "%CHROM",
                    "\t",
                    "%POS",
                    "\t",
                    "%REF",
                    "\t",
                    "%ALT",
                    ["\t", "%SAMPLE", "=", "%GT"],
                    "\n",
                ],
            ),
            (
                r"%CHROM\t%POS\t%REF\t%ALT[\t%SAMPLE=%GT{0}]\n",
                [
                    "%CHROM",
                    "\t",
                    "%POS",
                    "\t",
                    "%REF",
                    "\t",
                    "%ALT",
                    ["\t", "%SAMPLE", "=", ["%GT", 0]],
                    "\n",
                ],
            ),
            (
                r"GQ:[ %GQ] \t GT:[ %GT]\n",
                ["GQ:", [" ", "%GQ"], " ", "\t", " ", "GT:", [" ", "%GT"], "\n"],
            ),
            (
                r"[%SAMPLE %GT %DP\n]",
                [["%SAMPLE", " ", "%GT", " ", "%DP", "\n"]],
            ),
        ],
    )
    def test_valid_expressions(self, parser, expression, expected_result):
        assert parser(expression).as_list() == expected_result

    @pytest.mark.parametrize(
        "expression",
        [
            "%ac",
            "%AC {1}",
            "% CHROM",
        ],
    )
    def test_invalid_expressions(self, parser, expression):
        with pytest.raises(pp.ParseException):
            parser(expression)


class TestQueryFormatEvaluator:
    @pytest.fixture()
    def root(self):
        vcf_path = pathlib.Path("tests/data/vcf/sample.vcf.gz")
        vcz_path = vcz_path_cache(vcf_path)
        return zarr.open(vcz_path, mode="r")

    @pytest.mark.parametrize(
        ("query_format", "expected_result"),
        [
            (r"A\t", "A\t" * 9),
            (r"CHROM", "CHROM" * 9),
            (
                r"%CHROM:%POS\n",
                "19:111\n19:112\n20:14370\n20:17330\n20:1110696\n20:1230237\n20:1234567\n20:1235237\nX:10\n",
            ),
            (r"%INFO/DP\n", ".\n.\n14\n11\n10\n13\n9\n.\n.\n"),
            (r"%AC\n", ".\n.\n.\n.\n.\n.\n1,1\n.\n.\n"),
            (r"%AC{0}\n", ".\n.\n.\n.\n.\n.\n1\n.\n.\n"),
        ],
    )
    def test(self, root, query_format, expected_result):
        generator = QueryFormatGenerator(query_format)
        result = "".join(generator(root))
        assert result == expected_result

    @pytest.mark.parametrize(
        ("query_format", "expected_result"),
        [(r"%QUAL\n", "9.6\n10\n29\n3\n67\n47\n50\n.\n10\n")],
    )
    def test_with_parse_results(self, root, query_format, expected_result):
        parser = QueryFormatParser()
        parse_results = parser(query_format)
        generator = QueryFormatGenerator(parse_results)
        result = "".join(generator(root))
        assert result == expected_result


def test_write_query__include_exclude(tmp_path):
    original = pathlib.Path("tests/data/vcf") / "sample.vcf.gz"
    vcz = vcz_path_cache(original)
    output = tmp_path.joinpath("output.vcf")

    query_format = r"%POS\n"
    variant_site_filter = "POS > 1"

    with pytest.raises(
        ValueError,
        match=re.escape(
            "Cannot handle both an include expression and an exclude expression."
        ),
    ):
        write_query(
            vcz,
            output,
            query_format=query_format,
            include=variant_site_filter,
            exclude=variant_site_filter,
        )



import pathlib
import subprocess

import click.testing as ct
import pytest

import vcztools.cli as cli

from .utils import assert_vcfs_close, vcz_path_cache


def run_bcftools(args: str) -> str:
    """Run bcftools (which must be on the PATH) and return stdout as a string."""
    completed = subprocess.run(
        f"bcftools {args}", capture_output=True, check=True, shell=True
    )
    return completed.stdout.decode("utf-8")


def run_vcztools(args: str) -> str:
    """Run run_vcztools and return stdout as a string."""
    runner = ct.CliRunner(mix_stderr=False)
    result = runner.invoke(
        cli.vcztools_main,
        args,
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    return result.stdout


# fmt: off
@pytest.mark.parametrize(
    ("args", "vcf_file"),
    [
        ("view --no-version", "sample.vcf.gz"),
        ("view --no-version", "chr22.vcf.gz"),
        ("view --no-version -i 'INFO/DP > 10'", "sample.vcf.gz"),
        ("view --no-version -i 'FMT/DP >= 5 && FMT/GQ > 10'", "sample.vcf.gz"),
        ("view --no-version -i 'FMT/DP >= 5 & FMT/GQ>10'", "sample.vcf.gz"),
        (
                "view --no-version -i '(QUAL > 10 || FMT/GQ>10) && POS > 100000'",
                "sample.vcf.gz"
        ),
        (
                "view --no-version -i '(FMT/DP >= 8 | FMT/GQ>40) && POS > 100000'",
                "sample.vcf.gz"
        ),
        (
                "view --no-version -e '(FMT/DP >= 8 | FMT/GQ>40) && POS > 100000'",
                "sample.vcf.gz"
        ),
        (
                "view --no-version -G",
                "sample.vcf.gz"
        ),
        # Temporarily removing until sample handling fixed:
        # https://github.com/sgkit-dev/vcztools/issues/121
        # (
        #         "view --no-update --no-version --samples-file "
        #         "tests/data/txt/samples.txt",
        #         "sample.vcf.gz"),
        # ("view -I --no-version -S tests/data/txt/samples.txt", "sample.vcf.gz"),
        # ("view --no-version -s NA00001", "sample.vcf.gz"),
        # ("view --no-version -s NA00001,NA00003", "sample.vcf.gz"),
        # ("view --no-version -s HG00096", "1kg_2020_chrM.vcf.gz"),
        # ("view --no-version -s '' --force-samples", "sample.vcf.gz"),
        # ("view --no-version -s ^NA00001", "sample.vcf.gz"),
        # ("view --no-version -s ^NA00003,NA00002", "sample.vcf.gz"),
        # ("view --no-version -s ^NA00003,NA00002,NA00003", "sample.vcf.gz"),
        # ("view --no-version -S ^tests/data/txt/samples.txt", "sample.vcf.gz"),
    ]
)
# fmt: on
def test_vcf_output(tmp_path, args, vcf_file):
    original = pathlib.Path("tests/data/vcf") / vcf_file
    vcz = vcz_path_cache(original)

    bcftools_out = run_bcftools(f"{args} {original}")
    bcftools_out_file = tmp_path.joinpath("bcftools_out.vcf")
    with open(bcftools_out_file, "w") as f:
        f.write(bcftools_out)

    vcztools_out = run_vcztools(f"{args} {vcz}")
    vcztools_out_file = tmp_path.joinpath("vcztools_out.vcf")
    with open(vcztools_out_file, "w") as f:
        f.write(vcztools_out)

    assert_vcfs_close(bcftools_out_file, vcztools_out_file)

@pytest.mark.parametrize(
    ("args", "vcf_file"),
    [
        ("view --no-version", "sample.vcf.gz")
    ],
)
def test_vcf_output_with_output_option(tmp_path, args, vcf_file):
    vcf_path = pathlib.Path("tests/data/vcf") / vcf_file
    vcz_path = vcz_path_cache(vcf_path)

    bcftools_out_file = tmp_path.joinpath("bcftools_out.vcf")
    vcztools_out_file = tmp_path.joinpath("vcztools_out.vcf")

    bcftools_args = f"{args} -o {bcftools_out_file}"
    vcztools_args = f"{args} -o {vcztools_out_file}"

    run_bcftools(f"{bcftools_args} {vcf_path}")
    run_vcztools(f"{vcztools_args} {vcz_path}")

    assert_vcfs_close(bcftools_out_file, vcztools_out_file)


@pytest.mark.parametrize(
    ("args", "vcf_name"),
    [
        ("index -n", "sample.vcf.gz"),
        ("index --nrecords", "1kg_2020_chrM.vcf.gz"),
        ("index -s", "sample.vcf.gz"),
        ("index --stats", "1kg_2020_chrM.vcf.gz"),
        ("query -l", "sample.vcf.gz"),
        ("query --list-samples", "1kg_2020_chrM.vcf.gz"),
        (r"query -f 'A\n'", "sample.vcf.gz"),
        (r"query -f '%CHROM:%POS\n'", "sample.vcf.gz"),
        (r"query -f '%INFO/DP\n'", "sample.vcf.gz"),
        (r"query -f '%AC{0}\n'", "sample.vcf.gz"),
        (r"query -f '%REF\t%ALT\n'", "sample.vcf.gz"),
        (r"query -f '%ALT{1}\n'", "sample.vcf.gz"),
        (r"query -f '%ID\n'", "sample.vcf.gz"),
        (r"query -f '%QUAL\n'", "sample.vcf.gz"),
        (r"query -f '%FILTER\n'", "sample.vcf.gz"),
        (r"query --format '%FILTER\n'", "1kg_2020_chrM.vcf.gz"),
        (r"query -f '%POS\n' -i 'POS=112'", "sample.vcf.gz"),
        (r"query -f '%POS\n' -e 'POS=112'", "sample.vcf.gz"),
        (r"query -f '[%CHROM\t]\n'", "sample.vcf.gz"),
        (r"query -f '[%CHROM\t]\n' -i 'POS=112'", "sample.vcf.gz"),
        (r"query -f '%CHROM\t%POS\t%REF\t%ALT[\t%SAMPLE=%GT]\n'", "sample.vcf.gz"),
        (r"query -f 'GQ:[ %GQ] \t GT:[ %GT]\n'", "sample.vcf.gz"),
        (r"query -f '[%CHROM:%POS %SAMPLE %GT\n]'", "sample.vcf.gz"),
        (r"query -f '[%SAMPLE %GT %DP\n]'", "sample.vcf.gz"),
    ],
)
def test_output(tmp_path, args, vcf_name):
    vcf_path = pathlib.Path("tests/data/vcf") / vcf_name
    vcz_path = vcz_path_cache(vcf_path)

    bcftools_output = run_bcftools(f"{args} {vcf_path}")
    vcztools_output = run_vcztools(f"{args} {vcz_path}")

    assert vcztools_output == bcftools_output



from typing import Optional

import pytest

from vcztools.regions import parse_region_string


@pytest.mark.parametrize(
    ("targets", "expected"),
    [
        ("chr1", ("chr1", None, None)),
        ("chr1:12", ("chr1", 12, 12)),
        ("chr1:12-", ("chr1", 12, None)),
        ("chr1:12-103", ("chr1", 12, 103)),
    ],
)
def test_parse_region_string(
    targets: str, expected: tuple[str, Optional[int], Optional[int]]
):
    assert parse_region_string(targets) == expected



import pathlib

import pytest

from tests.utils import vcz_path_cache
from vcztools.vcf_writer import write_vcf

from .utils import assert_vcfs_close


@pytest.mark.parametrize(
    "vcf_file",
    [
        "sample.vcf.gz",
        "1kg_2020_chr20_annotations.bcf",
        "1kg_2020_chrM.vcf.gz",
        "field_type_combos.vcf.gz",
    ],
)
def test_vcf_to_zarr_to_vcf__real_files(tmp_path, vcf_file):
    original = pathlib.Path("tests/data/vcf") / vcf_file
    vcz = vcz_path_cache(original)
    generated = tmp_path.joinpath("output.vcf")
    write_vcf(vcz, generated, no_version=True)
    assert_vcfs_close(original, generated)



import concurrent.futures
import functools
import io
import re
import sys
from datetime import datetime
from typing import Optional

import numpy as np
import zarr

from vcztools.regions import (
    parse_regions,
    parse_targets,
    regions_to_chunk_indexes,
    regions_to_selection,
)
from vcztools.utils import (
    open_file_like,
    search,
)

from . import _vcztools, constants
from .constants import RESERVED_VARIABLE_NAMES
from .filter import FilterExpressionEvaluator, FilterExpressionParser

# references to the VCF spec are for https://samtools.github.io/hts-specs/VCFv4.3.pdf

# [Table 1: Reserved INFO keys]
RESERVED_INFO_KEY_DESCRIPTIONS = {
    "AA": "Ancestral allele",
    "AC": "Allele count in genotypes, for each ALT allele, in the same order as listed",
    "AD": "Total read depth for each allele",
    "ADF": "Read depth for each allele on the forward strand",
    "ADR": "Read depth for each allele on the reverse strand",
    "AF": "Allele frequency for each ALT allele in the same order as listed",
    "AN": "Total number of alleles in called genotypes",
    "BQ": "RMS base quality",
    "CIGAR": "Cigar string describing how to align an alternate allele to the reference"
    "allele",
    "DB": "dbSNP membership",
    "DP": "Combined depth across samples",
    "END": "End position on CHROM",
    "H2": "HapMap2 membership",
    "H3": "HapMap3 membership",
    "MQ": "RMS mapping quality",
    "MQ0": "Number of MAPQ == 0 reads",
    "NS": "Number of samples with data",
    "SB": "Strand bias",
    "SOMATIC": "Somatic mutation",
    "VALIDATED": "Validated by follow-up experiment",
    "1000G": "1000 Genomes membership",
}

# [Table 2: Reserved genotype keys]
RESERVED_FORMAT_KEY_DESCRIPTIONS = {
    "AD": "Read depth for each allele",
    "ADF": "Read depth for each allele on the forward strand",
    "ADR": "Read depth for each allele on the reverse strand",
    "DP": "Read depth",
    "EC": "Expected alternate allele counts",
    "FT": 'Filter indicating if this genotype was "called"',
    "GL": "Genotype likelihoods",
    "GP": "Genotype posterior probabilities",
    "GQ": "Conditional genotype quality",
    "GT": "Genotype",
    "HQ": "Haplotype quality",
    "MQ": "RMS mapping quality",
    "PL": "Phred-scaled genotype likelihoods rounded to the closest integer",
    "PP": "Phred-scaled genotype posterior probabilities rounded to the closest "
    "integer",
    "PQ": "Phasing quality",
    "PS": "Phase set",
}


def dims(arr):
    return arr.attrs["_ARRAY_DIMENSIONS"]


def write_vcf(
    vcz,
    output,
    *,
    header_only: bool = False,
    no_header: bool = False,
    no_version: bool = False,
    variant_regions=None,
    variant_targets=None,
    no_update=None,
    samples=None,
    drop_genotypes: bool = False,
    include: Optional[str] = None,
    exclude: Optional[str] = None,
) -> None:
    """Convert a dataset to a VCF file.

    If a VCF header is required and ``vcf_header`` attribute is present in the dataset,
    it will be used to generate the new VCF header. In this case, any variables in the
    dataset that are not specified in this header will have corresponding header lines
    added, and any lines in the header without a corresponding variable in the dataset
    will be omitted.

    In the case of no ``vcf_header`` attribute, a VCF header will
    be generated, and will include all variables in the dataset.

    Float fields are written with up to 3 decimal places of precision.
    Exponent/scientific notation is *not* supported, so values less than
    ``5e-4`` will be rounded to zero.

    Data is written sequentially to VCF, using C to optimize the write
    throughput speed.

    Data is loaded into memory in chunks sized according to the chunking along
    the variants dimension. Chunking in other dimensions (such as samples) is
    ignored for the purposes of writing VCF. If the dataset is not chunked
    (because it does not originate from Zarr or Dask, for example), then it
    will all be loaded into memory at once.

    The output is *not* compressed or indexed. It is therefore recommended to
    post-process the output using external tools such as ``bgzip(1)``,
    ``bcftools(1)``, or ``tabix(1)``.

    Parameters
    ----------
    input
        Dataset to convert to VCF.
    output
        A path or text file object that the output VCF should be written to.
    """

    root = zarr.open(vcz, mode="r")

    with open_file_like(output) as output:
        if samples and drop_genotypes:
            raise ValueError("Cannot select samples and drop genotypes.")
        elif drop_genotypes:
            sample_ids = []
            samples_selection = np.array([])
        elif samples is None:
            sample_ids = root["sample_id"][:]
            samples_selection = None
        else:
            all_samples = root["sample_id"][:]
            exclude_samples = samples.startswith("^")
            samples = samples.lstrip("^")
            sample_ids = np.array(samples.split(","))
            if np.all(sample_ids == np.array("")):
                sample_ids = np.empty((0,))

            samples_selection = search(all_samples, sample_ids)
            if exclude_samples:
                samples_selection = np.setdiff1d(
                    np.arange(all_samples.size), samples_selection
                )
                sample_ids = all_samples[samples_selection]

        if not no_header:
            original_header = root.attrs.get("vcf_header", None)
            vcf_header = _generate_header(
                root, original_header, sample_ids, no_version=no_version
            )
            print(vcf_header, end="", file=output)

        if header_only:
            return

        pos = root["variant_position"]
        num_variants = pos.shape[0]

        if num_variants == 0:
            return

        contigs = root["contig_id"][:].astype("S")
        filters = root["filter_id"][:].astype("S")

        if include and exclude:
            raise ValueError(
                "Cannot handle both an include expression and an exclude expression."
            )
        elif include or exclude:
            parser = FilterExpressionParser()
            parse_results = parser(include or exclude)[0]
            filter_evaluator = FilterExpressionEvaluator(
                parse_results, invert=bool(exclude)
            )
            filter_evaluator = functools.partial(filter_evaluator, root)
        else:
            filter_evaluator = None

        if variant_regions is None and variant_targets is None:
            # no regions or targets selected
            with concurrent.futures.ThreadPoolExecutor() as executor:
                preceding_future = None
                for v_chunk in range(pos.cdata_shape[0]):
                    v_mask_chunk = (
                        filter_evaluator(v_chunk) if filter_evaluator else None
                    )
                    future = executor.submit(
                        c_chunk_to_vcf,
                        root,
                        v_chunk,
                        v_mask_chunk,
                        samples_selection,
                        contigs,
                        filters,
                        output,
                        drop_genotypes=drop_genotypes,
                        no_update=no_update,
                        preceding_future=preceding_future,
                    )
                    if preceding_future:
                        concurrent.futures.wait((preceding_future,))
                    preceding_future = future
        else:
            contigs_u = root["contig_id"][:].astype("U").tolist()
            regions = parse_regions(variant_regions, contigs_u)
            targets, complement = parse_targets(variant_targets, contigs_u)

            # Use the region index to find the chunks that overlap specfied regions or
            # targets
            region_index = root["region_index"][:]
            chunk_indexes = regions_to_chunk_indexes(
                regions,
                targets,
                complement,
                region_index,
            )

            # Then use only load required variant_contig/position chunks
            if len(chunk_indexes) == 0:
                # no chunks - no variants to write
                return
            elif len(chunk_indexes) == 1:
                # single chunk
                block_sel = chunk_indexes[0]
            else:
                # zarr.blocks doesn't support int array indexing - use that when it does
                block_sel = slice(chunk_indexes[0], chunk_indexes[-1] + 1)

            region_variant_contig = root["variant_contig"].blocks[block_sel][:]
            region_variant_position = root["variant_position"].blocks[block_sel][:]
            region_variant_length = root["variant_length"].blocks[block_sel][:]

            # Find the final variant selection
            variant_selection = regions_to_selection(
                regions,
                targets,
                complement,
                region_variant_contig,
                region_variant_position,
                region_variant_length,
            )
            variant_mask = np.zeros(region_variant_position.shape[0], dtype=bool)
            variant_mask[variant_selection] = 1
            # Use zarr arrays to get mask chunks aligned with the main data
            # for convenience.
            z_variant_mask = zarr.array(variant_mask, chunks=pos.chunks[0])
            with concurrent.futures.ThreadPoolExecutor() as executor:
                preceding_future = None
                for i, v_chunk in enumerate(chunk_indexes):
                    v_mask_chunk = z_variant_mask.blocks[i]

                    if filter_evaluator and np.any(v_mask_chunk):
                        v_mask_chunk = np.logical_and(
                            v_mask_chunk, filter_evaluator(v_chunk)
                        )
                    if np.any(v_mask_chunk):
                        future = executor.submit(
                            c_chunk_to_vcf,
                            root,
                            v_chunk,
                            v_mask_chunk,
                            samples_selection,
                            contigs,
                            filters,
                            output,
                            drop_genotypes=drop_genotypes,
                            no_update=no_update,
                            preceding_future=preceding_future,
                        )
                        if preceding_future:
                            concurrent.futures.wait((preceding_future,))
                        preceding_future = future


def get_vchunk_array(zarray, v_chunk, mask, samples_selection=None):
    v_chunksize = zarray.chunks[0]
    start = v_chunksize * v_chunk
    end = v_chunksize * (v_chunk + 1)
    if samples_selection is None:
        result = zarray[start:end]
    else:
        result = zarray.oindex[start:end, samples_selection]
    if mask is not None:
        result = result[mask]
    return result


def c_chunk_to_vcf(
    root,
    v_chunk,
    v_mask_chunk,
    samples_selection,
    contigs,
    filters,
    output,
    *,
    drop_genotypes,
    no_update,
    preceding_future=None,
):
    chrom = contigs[get_vchunk_array(root["variant_contig"], v_chunk, v_mask_chunk)]
    # TODO check we don't truncate silently by doing this
    pos = get_vchunk_array(root["variant_position"], v_chunk, v_mask_chunk).astype(
        np.int32
    )
    id = get_vchunk_array(root["variant_id"], v_chunk, v_mask_chunk).astype("S")
    alleles = get_vchunk_array(root["variant_allele"], v_chunk, v_mask_chunk)
    qual = get_vchunk_array(root["variant_quality"], v_chunk, v_mask_chunk)
    filter_ = get_vchunk_array(root["variant_filter"], v_chunk, v_mask_chunk)
    format_fields = {}
    info_fields = {}
    num_samples = len(samples_selection) if samples_selection is not None else None
    gt = None
    gt_phased = None

    if "call_genotype" in root and not drop_genotypes:
        if samples_selection is not None and num_samples != 0:
            gt = get_vchunk_array(
                root["call_genotype"], v_chunk, v_mask_chunk, samples_selection
            )
        else:
            gt = get_vchunk_array(root["call_genotype"], v_chunk, v_mask_chunk)

        if (
            "call_genotype_phased" in root
            and not drop_genotypes
            and (samples_selection is None or num_samples > 0)
        ):
            gt_phased = get_vchunk_array(
                root["call_genotype_phased"],
                v_chunk,
                v_mask_chunk,
                samples_selection,
            )
        else:
            gt_phased = np.zeros_like(gt, dtype=bool)

    for name, zarray in root.arrays():
        if (
            name.startswith("call_")
            and not name.startswith("call_genotype")
            and num_samples != 0
        ):
            vcf_name = name[len("call_") :]
            format_fields[vcf_name] = get_vchunk_array(
                zarray, v_chunk, v_mask_chunk, samples_selection
            )
            if num_samples is None:
                num_samples = zarray.shape[1]
        elif name.startswith("variant_") and name not in RESERVED_VARIABLE_NAMES:
            vcf_name = name[len("variant_") :]
            info_fields[vcf_name] = get_vchunk_array(zarray, v_chunk, v_mask_chunk)

    ref = alleles[:, 0].astype("S")
    alt = alleles[:, 1:].astype("S")

    if len(id.shape) == 1:
        id = id.reshape((-1, 1))
    if (
        not no_update
        and samples_selection is not None
        and "call_genotype" in root
        and not drop_genotypes
    ):
        # Recompute INFO/AC and INFO/AN
        info_fields |= _compute_info_fields(gt, alt)
    if num_samples == 0:
        gt = None
    if gt is not None and num_samples is None:
        num_samples = gt.shape[1]

    num_variants = len(pos)
    encoder = _vcztools.VcfEncoder(
        num_variants,
        num_samples if num_samples is not None else 0,
        chrom=chrom,
        pos=pos,
        id=id,
        alt=alt,
        ref=ref,
        qual=qual,
        filter_ids=filters,
        filter=filter_,
    )
    # print(encoder.arrays)
    if gt is not None:
        encoder.add_gt_field(gt, gt_phased)
    for name, zarray in info_fields.items():
        # print(array.dtype.kind)
        if zarray.dtype.kind in ("O", "U"):
            zarray = zarray.astype("S")
        if len(zarray.shape) == 1:
            zarray = zarray.reshape((num_variants, 1))
        encoder.add_info_field(name, zarray)

    if num_samples != 0:
        for name, zarray in format_fields.items():
            if zarray.dtype.kind in ("O", "U"):
                zarray = zarray.astype("S")
            if len(zarray.shape) == 2:
                zarray = zarray.reshape((num_variants, num_samples, 1))
            encoder.add_format_field(name, zarray)

    if preceding_future:
        concurrent.futures.wait((preceding_future,))

    # TODO: (1) make a guess at this based on number of fields and samples,
    # and (2) log a DEBUG message when we have to double.
    buflen = 1024
    for j in range(num_variants):
        failed = True
        while failed:
            try:
                line = encoder.encode(j, buflen)
                failed = False
            except _vcztools.VczBufferTooSmall:
                buflen *= 2
                # print("Bumping buflen to", buflen)
        print(line, file=output)


def _generate_header(ds, original_header, sample_ids, *, no_version: bool = False):
    output = io.StringIO()

    contigs = list(ds["contig_id"][:])
    filters = list(ds["filter_id"][:])
    info_fields = []
    format_fields = []

    if "call_genotype" in ds and len(sample_ids) > 0:
        # GT must be the first field if present, per the spec (section 1.6.2)
        format_fields.append("GT")

    for var in sorted(ds.keys()):
        arr = ds[var]
        if (
            var.startswith("variant_")
            and not var.endswith("_fill")
            and not var.endswith("_mask")
            and var not in RESERVED_VARIABLE_NAMES
            and dims(arr)[0] == "variants"
        ):
            key = var[len("variant_") :]
            info_fields.append(key)
        elif (
            len(sample_ids) > 0
            and var.startswith("call_")
            and not var.endswith("_fill")
            and not var.endswith("_mask")
            and dims(arr)[0] == "variants"
            and dims(arr)[1] == "samples"
        ):
            key = var[len("call_") :]
            if key in ("genotype", "genotype_phased"):
                continue
            format_fields.append(key)

    if original_header is None:  # generate entire header
        # [1.4.1 File format]
        print("##fileformat=VCFv4.3", file=output)

        if "source" in ds.attrs:
            print(f'##source={ds.attrs["source"]}', file=output)

    else:  # use original header fields where appropriate
        unstructured_pattern = re.compile("##([^=]+)=([^<].*)")
        structured_pattern = re.compile("##([^=]+)=(<.*)")

        for line in original_header.split("\n"):
            if re.fullmatch(unstructured_pattern, line):
                print(line, file=output)
            else:
                match = re.fullmatch(structured_pattern, line)
                if match:
                    category = match.group(1)
                    id_pattern = re.compile("ID=([^,>]+)")
                    key = id_pattern.findall(line)[0]
                    if category not in ("contig", "FILTER", "INFO", "FORMAT"):
                        # output other structured fields
                        print(line, file=output)
                    # only output certain categories if in dataset
                    elif category == "contig" and key in contigs:
                        contigs.remove(key)
                        print(line, file=output)
                    elif category == "FILTER" and key in filters:
                        filters.remove(key)
                        print(line, file=output)
                    elif category == "INFO" and key in info_fields:
                        info_fields.remove(key)
                        print(line, file=output)
                    elif category == "FORMAT" and key in format_fields:
                        format_fields.remove(key)
                        print(line, file=output)

    # add all fields that are not in the original header
    # or all fields if there was no original header

    # [1.4.2 Information field format]
    for key in info_fields:
        arr = ds[f"variant_{key}"]
        category = "INFO"
        vcf_number = _array_to_vcf_number(category, key, arr)
        vcf_type = _array_to_vcf_type(arr)
        vcf_description = arr.attrs.get(
            "description", RESERVED_INFO_KEY_DESCRIPTIONS.get(key, "")
        )
        print(
            f'##INFO=<ID={key},Number={vcf_number},Type={vcf_type},Description="{vcf_description}">',
            file=output,
        )

    # [1.4.3 Filter field format]
    for filter in filters:
        print(f'##FILTER=<ID={filter},Description="">', file=output)

    # [1.4.4 Individual format field format]
    for key in format_fields:
        if key == "GT":
            print(
                '##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">',
                file=output,
            )
        else:
            arr = ds[f"call_{key}"]
            category = "FORMAT"
            vcf_number = _array_to_vcf_number(category, key, arr)
            vcf_type = _array_to_vcf_type(arr)
            vcf_description = arr.attrs.get(
                "description", RESERVED_FORMAT_KEY_DESCRIPTIONS.get(key, "")
            )
            print(
                f'##FORMAT=<ID={key},Number={vcf_number},Type={vcf_type},Description="{vcf_description}">',
                file=output,
            )

    # [1.4.7 Contig field format]
    contig_lengths = (
        ds.attrs["contig_lengths"] if "contig_lengths" in ds.attrs else None
    )
    for i, contig in enumerate(contigs):
        if contig_lengths is None:
            print(f"##contig=<ID={contig}>", file=output)
        else:
            print(f"##contig=<ID={contig},length={contig_lengths[i]}>", file=output)

    if not no_version:
        print(
            f"##vcztools_viewCommand={' '.join(sys.argv[1:])}; Date={datetime.now()}",
            file=output,
        )

    # [1.5 Header line syntax]
    print(
        "#CHROM",
        "POS",
        "ID",
        "REF",
        "ALT",
        "QUAL",
        "FILTER",
        "INFO",
        sep="\t",
        end="",
        file=output,
    )

    if len(sample_ids) > 0:
        print(end="\t", file=output)
        print("FORMAT", *sample_ids, sep="\t", file=output)
    else:
        print(file=output)

    return output.getvalue()


def _array_to_vcf_number(category, key, a):
    # reverse of vcf_number_to_dimension_and_size
    if a.dtype == bool:
        return 0
    elif category == "INFO" and len(dims(a)) == 1:
        return 1
    elif category == "FORMAT" and len(dims(a)) == 2:
        return 1

    last_dim = dims(a)[-1]
    if last_dim == "alt_alleles":
        return "A"
    elif last_dim == "alleles":
        return "R"
    elif last_dim == "genotypes":
        return "G"
    elif last_dim == f"{category}_{key}_dim":
        return a.shape[-1]
    else:
        raise ValueError(
            f"Cannot determine VCF Number for dimension name '{last_dim}' in {a}"
        )


def _array_to_vcf_type(a):
    # reverse of _vcf_type_to_numpy
    if a.dtype == bool:
        return "Flag"
    elif np.issubdtype(a.dtype, np.integer):
        return "Integer"
    elif np.issubdtype(a.dtype, np.float32):
        return "Float"
    elif a.dtype.str == "|S1":
        return "Character"
    elif a.dtype.kind in ("O", "S", "U"):
        return "String"
    else:
        raise ValueError(f"Unsupported dtype: {a.dtype}")


def _compute_info_fields(gt: np.ndarray, alt: np.ndarray):
    flatter_gt = gt.reshape((gt.shape[0], -1))
    allele_count = alt.shape[1] + 1

    def filter_and_bincount(values: np.ndarray):
        positive = values[values > 0]
        return np.bincount(positive, minlength=allele_count)[1:]

    computed_ac = np.apply_along_axis(filter_and_bincount, 1, flatter_gt).astype(
        np.int32
    )
    computed_ac[alt == b""] = constants.INT_FILL
    computed_an = np.sum(flatter_gt >= 0, axis=1, dtype=np.int32)

    return {
        "AC": computed_ac,
        "AN": computed_an,
    }



import functools
import itertools
import math
from typing import Callable, Optional, Union

import numpy as np
import pyparsing as pp
import zarr

from vcztools import constants
from vcztools.filter import FilterExpressionEvaluator, FilterExpressionParser
from vcztools.utils import vcf_name_to_vcz_name


def list_samples(vcz_path, output):
    root = zarr.open(vcz_path, mode="r")

    sample_ids = root["sample_id"][:]
    print("\n".join(sample_ids), file=output)


class QueryFormatParser:
    def __init__(self):
        info_tag_pattern = pp.Combine(
            pp.Literal("%INFO/") + pp.Word(pp.srange("[A-Z]"))
        )
        tag_pattern = info_tag_pattern | pp.Combine(
            pp.Literal("%") + pp.Regex(r"[A-Z]+\d?")
        )
        subfield_pattern = pp.Group(
            tag_pattern
            + pp.Literal("{").suppress()
            + pp.common.integer
            + pp.Literal("}").suppress()
        ).set_results_name("subfield")
        newline_pattern = pp.Literal("\\n").set_parse_action(pp.replace_with("\n"))
        tab_pattern = pp.Literal("\\t").set_parse_action(pp.replace_with("\t"))
        format_pattern = pp.Forward()
        sample_loop_pattern = pp.Group(
            pp.Literal("[").suppress() + format_pattern + pp.Literal("]").suppress()
        ).set_results_name("sample loop")
        format_pattern <<= pp.ZeroOrMore(
            sample_loop_pattern
            | subfield_pattern
            | tag_pattern
            | newline_pattern
            | tab_pattern
            | pp.White()
            | pp.Word(pp.printables, exclude_chars=r"\{}[]%")
        ).leave_whitespace()

        self._parser = functools.partial(format_pattern.parse_string, parse_all=True)

    def __call__(self, *args, **kwargs):
        assert len(args) == 1
        assert not kwargs

        return self._parser(args[0])


class QueryFormatGenerator:
    def __init__(
        self,
        query_format: Union[str, pp.ParseResults],
        *,
        include: Optional[str] = None,
        exclude: Optional[str] = None,
    ):
        if isinstance(query_format, str):
            parser = QueryFormatParser()
            parse_results = parser(query_format)
        else:
            assert isinstance(query_format, pp.ParseResults)
            parse_results = query_format

        self._generator = self._compose_generator(
            parse_results, include=include, exclude=exclude
        )

    def __call__(self, *args, **kwargs):
        assert len(args) == 1
        assert not kwargs

        yield from self._generator(args[0])

    def _compose_gt_generator(self) -> Callable:
        def generate(root):
            gt_zarray = root["call_genotype"]
            v_chunk_size = gt_zarray.chunks[0]

            if "call_genotype_phased" in root:
                phase_zarray = root["call_genotype_phased"]
                assert gt_zarray.chunks[:2] == phase_zarray.chunks
                assert gt_zarray.shape[:2] == phase_zarray.shape

                for v_chunk_index in range(gt_zarray.cdata_shape[0]):
                    start = v_chunk_index * v_chunk_size
                    end = start + v_chunk_size

                    for gt_row, phase in zip(
                        gt_zarray[start:end], phase_zarray[start:end]
                    ):

                        def stringify(gt_and_phase: tuple):
                            gt, phase = gt_and_phase
                            gt = [
                                str(allele) if allele != constants.INT_MISSING else "."
                                for allele in gt
                                if allele != constants.INT_FILL
                            ]
                            separator = "|" if phase else "/"
                            return separator.join(gt)

                        gt_row = gt_row.tolist()
                        yield map(stringify, zip(gt_row, phase))
            else:
                # TODO: Support datasets without the phasing data
                raise NotImplementedError

        return generate

    def _compose_sample_ids_generator(self) -> Callable:
        def generate(root):
            variant_count = root["variant_position"].shape[0]
            sample_ids = root["sample_id"][:].tolist()
            yield from itertools.repeat(sample_ids, variant_count)

        return generate

    def _compose_tag_generator(
        self, tag: str, *, subfield=False, sample_loop=False
    ) -> Callable:
        assert tag.startswith("%")
        tag = tag[1:]

        if tag == "GT":
            return self._compose_gt_generator()

        if tag == "SAMPLE":
            return self._compose_sample_ids_generator()

        def generate(root):
            vcz_names = set(root.keys())
            vcz_name = vcf_name_to_vcz_name(vcz_names, tag)
            zarray = root[vcz_name]
            contig_ids = root["contig_id"][:] if tag == "CHROM" else None
            filter_ids = root["filter_id"][:] if tag == "FILTER" else None
            v_chunk_size = zarray.chunks[0]

            for v_chunk_index in range(zarray.cdata_shape[0]):
                start = v_chunk_index * v_chunk_size
                end = start + v_chunk_size

                for row in zarray[start:end]:
                    is_missing = np.any(row == -1)

                    if tag == "CHROM":
                        assert contig_ids is not None
                        row = contig_ids[row]
                    if tag == "REF":
                        row = row[0]
                    if tag == "ALT":
                        row = [allele for allele in row[1:] if allele] or "."
                    if tag == "FILTER":
                        assert filter_ids is not None

                        if np.any(row):
                            row = filter_ids[row]
                        else:
                            row = "."
                    if tag == "QUAL":
                        if math.isnan(row):
                            row = "."
                        else:
                            row = f"{row:g}"
                    if (
                        not subfield
                        and not sample_loop
                        and (isinstance(row, np.ndarray) or isinstance(row, list))
                    ):
                        row = ",".join(map(str, row))

                    if sample_loop:
                        sample_count = root["sample_id"].shape[0]

                        if isinstance(row, np.ndarray):
                            row = row.tolist()
                            row = [
                                str(element)
                                if element != constants.INT_MISSING
                                else "."
                                for element in row
                                if element != constants.INT_FILL
                            ]
                            yield row
                        else:
                            yield itertools.repeat(str(row), sample_count)
                    else:
                        yield row if not is_missing else "."

        return generate

    def _compose_subfield_generator(self, parse_results: pp.ParseResults) -> Callable:
        assert len(parse_results) == 2

        tag, subfield_index = parse_results
        tag_generator = self._compose_tag_generator(tag, subfield=True)

        def generate(root):
            for tag in tag_generator(root):
                if isinstance(tag, str):
                    assert tag == "."
                    yield "."
                else:
                    if subfield_index < len(tag):
                        yield tag[subfield_index]
                    else:
                        yield "."

        return generate

    def _compose_sample_loop_generator(
        self, parse_results: pp.ParseResults
    ) -> Callable:
        generators = map(
            functools.partial(self._compose_element_generator, sample_loop=True),
            parse_results,
        )

        def generate(root):
            iterables = (generator(root) for generator in generators)
            zipped = zip(*iterables)
            zipped_zipped = (zip(*element) for element in zipped)
            flattened_zipped_zipped = (
                (
                    subsubelement
                    for subelement in element  # sample-wise
                    for subsubelement in subelement
                )
                for element in zipped_zipped  # variant-wise
            )
            yield from map("".join, flattened_zipped_zipped)

        return generate

    def _compose_element_generator(
        self, element: Union[str, pp.ParseResults], *, sample_loop=False
    ) -> Callable:
        if isinstance(element, pp.ParseResults):
            if element.get_name() == "subfield":
                return self._compose_subfield_generator(element)
            elif element.get_name() == "sample loop":
                return self._compose_sample_loop_generator(element)

        assert isinstance(element, str)

        if element.startswith("%"):
            return self._compose_tag_generator(element, sample_loop=sample_loop)
        else:

            def generate(root):
                nonlocal element
                variant_count = root["variant_position"].shape[0]
                if sample_loop:
                    sample_count = root["sample_id"].shape[0]
                    for _ in range(variant_count):
                        yield itertools.repeat(element, sample_count)
                else:
                    yield from itertools.repeat(element, variant_count)

            return generate

    def _compose_filter_generator(
        self, *, include: Optional[str] = None, exclude: Optional[str] = None
    ) -> Callable:
        assert not (include and exclude)

        if not include and not exclude:

            def generate(root):
                variant_count = root["variant_position"].shape[0]
                yield from itertools.repeat(True, variant_count)

            return generate

        parser = FilterExpressionParser()
        parse_results = parser(include or exclude)[0]
        filter_evaluator = FilterExpressionEvaluator(
            parse_results, invert=bool(exclude)
        )

        def generate(root):
            nonlocal filter_evaluator

            filter_evaluator = functools.partial(filter_evaluator, root)
            variant_chunk_count = root["variant_position"].cdata_shape[0]

            for variant_chunk_index in range(variant_chunk_count):
                yield from filter_evaluator(variant_chunk_index)

        return generate

    def _compose_generator(
        self,
        parse_results: pp.ParseResults,
        *,
        include: Optional[str] = None,
        exclude: Optional[str] = None,
    ) -> Callable:
        generators = (
            self._compose_element_generator(element) for element in parse_results
        )
        filter_generator = self._compose_filter_generator(
            include=include, exclude=exclude
        )

        def generate(root) -> str:
            iterables = (generator(root) for generator in generators)
            filter_iterable = filter_generator(root)
            for results, filter_indicator in zip(zip(*iterables), filter_iterable):
                if filter_indicator:
                    results = map(str, results)
                    yield "".join(results)

        return generate


def write_query(
    vcz,
    output,
    *,
    query_format: str,
    include: Optional[str] = None,
    exclude: Optional[str] = None,
):
    if include and exclude:
        raise ValueError(
            "Cannot handle both an include expression and an exclude expression."
        )

    root = zarr.open(vcz, mode="r")
    generator = QueryFormatGenerator(query_format, include=include, exclude=exclude)

    for result in generator(root):
        print(result, sep="", end="", file=output)



import pathlib
import re
from unittest import mock

import click.testing as ct
import pytest

import vcztools.cli as cli
from tests.test_bcftools_validation import run_vcztools
from tests.utils import vcz_path_cache


@pytest.fixture()
def vcz_path():
    vcf_path = pathlib.Path("tests/data/vcf/sample.vcf.gz")
    return vcz_path_cache(vcf_path)


def test_version_header(vcz_path):
    output = run_vcztools(f"view {vcz_path}")
    assert output.find("##vcztools_viewCommand=") >= 0
    assert output.find("Date=") >= 0


class TestOutput:
    def test_view_unsupported_output(self, tmp_path, vcz_path):
        bad_output = tmp_path / "output.vcf.gz"

        with pytest.raises(
            ValueError,
            match=re.escape(
                "Only uncompressed VCF output supported, suffix .gz not allowed"
            ),
        ):
            run_vcztools(f"view --no-version {vcz_path} -o {bad_output}")

    @pytest.mark.parametrize("suffix", ["gz", "bgz", "bcf"])
    def test_view_unsupported_output_suffix(self, tmp_path, vcz_path, suffix):
        bad_output = tmp_path / f"output.vcf.{suffix}"

        with pytest.raises(ValueError, match=f".{suffix} not allowed"):
            run_vcztools(f"view --no-version {vcz_path} -o {bad_output}")

    def test_view_good_path(self, tmp_path, vcz_path):
        output_path = tmp_path / "tmp.vcf"
        runner = ct.CliRunner(mix_stderr=False)
        result = runner.invoke(
            cli.vcztools_main,
            f"view --no-version {vcz_path} -o {output_path}",
            catch_exceptions=False,
        )
        assert result.exit_code == 0
        assert len(result.stdout) == 0
        assert output_path.exists()

    def test_view_write_directory(self, tmp_path, vcz_path):
        runner = ct.CliRunner(mix_stderr=False)
        result = runner.invoke(
            cli.vcztools_main,
            f"view --no-version {vcz_path} -o {tmp_path}",
            catch_exceptions=False,
        )
        assert result.exit_code == 1
        assert len(result.stdout) == 0
        assert "Is a directory" in result.stderr

    def test_view_write_pipe(self, tmp_path, vcz_path):
        runner = ct.CliRunner(mix_stderr=False)
        result = runner.invoke(
            cli.vcztools_main,
            f"view --no-version {vcz_path} -o {tmp_path}",
            catch_exceptions=False,
        )
        assert result.exit_code == 1
        assert len(result.stdout) == 0
        assert "Is a directory" in result.stderr


# Removing until we reimplement sample handling:
# https://github.com/sgkit-dev/vcztools/issues/121
# def test_excluding_and_including_samples(vcz_path):
#     samples_file_path = pathlib.Path("tests/data/txt/samples.txt")
#     error_message = re.escape("vcztools does not support combining -s and -S")

#     with pytest.raises(AssertionError, match=error_message):
#         run_vcztools(f"view {vcz_path} -s NA00001 -S ^{samples_file_path}")
#     with pytest.raises(AssertionError, match=error_message):
#         run_vcztools(f"view {vcz_path} -s ^NA00001 -S {samples_file_path}")


@mock.patch("sys.exit")
@mock.patch("os.dup2")
def test_broken_pipe(mocked_dup2, mocked_exit, tmp_path):
    with open(tmp_path / "tmp.txt", "w") as output:
        with cli.handle_broken_pipe(output):
            raise BrokenPipeError()
        mocked_dup2.assert_called_once()
        mocked_exit.assert_called_once_with(1)


class TestQuery:
    def test_format_required(self, vcz_path):
        runner = ct.CliRunner(mix_stderr=False)
        result = runner.invoke(
            cli.vcztools_main,
            f"query {vcz_path} ",
            catch_exceptions=False,
        )
        assert result.exit_code == 2
        assert len(result.stdout) == 0
        assert len(result.stderr) > 0

    def test_path_required(self):
        runner = ct.CliRunner(mix_stderr=False)
        result = runner.invoke(
            cli.vcztools_main,
            "query --format=POS ",
            catch_exceptions=False,
        )
        assert result.exit_code == 2
        assert len(result.stdout) == 0
        assert len(result.stderr) > 0

    def test_list(self, vcz_path):
        result = run_vcztools(f"query -l {vcz_path}")
        assert list(result.splitlines()) == ["NA00001", "NA00002", "NA00003"]

    def test_list_ignores_output(self, vcz_path, tmp_path):
        output = tmp_path / "tmp.txt"
        result = run_vcztools(f"query -l {vcz_path} -o {output}")
        assert list(result.splitlines()) == ["NA00001", "NA00002", "NA00003"]
        assert not output.exists()

    def test_output(self, vcz_path, tmp_path):
        output = tmp_path / "tmp.txt"
        result = run_vcztools(f"query -f '%POS\n' {vcz_path} -o {output}")
        assert list(result.splitlines()) == []
        assert output.exists()


class TestIndex:
    def test_stats(self, vcz_path):
        result = run_vcztools(f"index -s {vcz_path}")
        assert list(result.splitlines()) == ["19\t.\t2", "20\t.\t6", "X\t.\t1"]

    def test_nrecords(self, vcz_path):
        result = run_vcztools(f"index -n {vcz_path}")
        assert list(result.splitlines()) == ["9"]

    def test_stats_and_nrecords(self, vcz_path):
        runner = ct.CliRunner(mix_stderr=False)
        result = runner.invoke(
            cli.vcztools_main,
            f"index -ns {vcz_path}",
            catch_exceptions=False,
        )
        assert result.exit_code == 2
        assert len(result.stdout) == 0
        assert len(result.stderr) > 0
        assert "Expected only one of --stats or --nrecords options" in result.stderr

    def test_no_stats_or_nrecords(self, vcz_path):
        runner = ct.CliRunner(mix_stderr=False)
        result = runner.invoke(
            cli.vcztools_main,
            f"index {vcz_path}",
            catch_exceptions=False,
        )
        assert result.exit_code == 2
        assert len(result.stdout) == 0
        assert len(result.stderr) > 0
        assert "Error: Building region indexes is not supported" in result.stderr


def test_top_level():
    runner = ct.CliRunner(mix_stderr=False)
    result = runner.invoke(
        cli.vcztools_main,
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert len(result.stdout) > 0
    assert len(result.stderr) == 0



import functools
import operator
from typing import Callable

import numpy as np
import pyparsing as pp

from vcztools.utils import vcf_name_to_vcz_name


class FilterExpressionParser:
    def __init__(self):
        constant_pattern = pp.common.number | pp.QuotedString('"')
        standard_tag_pattern = pp.Word(pp.srange("[A-Z]"))

        info_tag_pattern = pp.Combine(pp.Literal("INFO/") + standard_tag_pattern)
        format_tag_pattern = pp.Combine(
            (pp.Literal("FORMAT/") | pp.Literal("FMT/")) + standard_tag_pattern
        )
        tag_pattern = info_tag_pattern | format_tag_pattern | standard_tag_pattern

        identifier_pattern = tag_pattern | constant_pattern
        self._identifier_parser = functools.partial(
            identifier_pattern.parse_string, parse_all=True
        )

        comparison_pattern = pp.Group(
            tag_pattern + pp.one_of("== = != > >= < <=") + constant_pattern
        ).set_results_name("comparison")

        parentheses_pattern = pp.Forward()
        and_pattern = pp.Forward()
        or_pattern = pp.Forward()

        parentheses_pattern <<= (
            pp.Suppress("(") + or_pattern + pp.Suppress(")") | comparison_pattern
        )
        and_pattern <<= (
            pp.Group(
                parentheses_pattern + (pp.Keyword("&&") | pp.Keyword("&")) + and_pattern
            ).set_results_name("and")
            | parentheses_pattern
        )
        or_pattern <<= (
            pp.Group(
                and_pattern + (pp.Keyword("||") | pp.Keyword("|")) + or_pattern
            ).set_results_name("or")
            | and_pattern
        )

        self._parser = functools.partial(or_pattern.parse_string, parse_all=True)

    def __call__(self, *args, **kwargs):
        assert args or kwargs

        if args:
            assert len(args) == 1
            assert not kwargs
            expression = args[0]
        else:
            assert len(kwargs) == 1
            assert "expression" in kwargs
            expression = kwargs["expression"]

        return self.parse(expression)

    def parse(self, expression: str):
        return self._parser(expression)


class FilterExpressionEvaluator:
    def __init__(self, parse_results: pp.ParseResults, *, invert=False):
        self._composers = {
            "comparison": self._compose_comparison_evaluator,
            "and": self._compose_and_evaluator,
            "or": self._compose_or_evaluator,
        }
        self._comparators = {
            "==": operator.eq,
            "=": operator.eq,
            "!=": operator.ne,
            ">": operator.gt,
            ">=": operator.ge,
            "<": operator.lt,
            "<=": operator.le,
        }
        base_evaluator = self._compose_evaluator(parse_results)

        def evaluator(root, variant_chunk_index: int) -> np.ndarray:
            base_array = base_evaluator(root, variant_chunk_index)
            return np.any(base_array, axis=tuple(range(1, base_array.ndim)))

        if invert:

            def invert_evaluator(root, variant_chunk_index: int) -> np.ndarray:
                return np.logical_not(evaluator(root, variant_chunk_index))

            self._evaluator = invert_evaluator
        else:
            self._evaluator = evaluator

    def __call__(self, *args, **kwargs):
        assert len(args) == 2
        assert not kwargs

        return self._evaluator(*args)

    def _compose_comparison_evaluator(self, parse_results: pp.ParseResults) -> Callable:
        assert len(parse_results) == 3

        comparator = parse_results[1]
        comparator = self._comparators[comparator]

        def evaluator(root, variant_chunk_index: int) -> np.ndarray:
            vcf_name = parse_results[0]
            vcz_names = set(root.keys())
            vcz_name = vcf_name_to_vcz_name(vcz_names, vcf_name)
            zarray = root[vcz_name]
            variant_chunk_len = zarray.chunks[0]
            start = variant_chunk_len * variant_chunk_index
            end = start + variant_chunk_len
            # We load all samples (regardless of sample filtering)
            # to match bcftools' behavior.
            array = zarray[start:end]
            array = comparator(array, parse_results[2])

            if array.ndim > 2:
                return np.any(array, axis=tuple(range(2, array.ndim)))
            else:
                return array

        return evaluator

    def _compose_and_evaluator(self, parse_results: pp.ParseResults) -> Callable:
        assert len(parse_results) == 3
        assert parse_results[1] in {"&", "&&"}

        left_evaluator = self._compose_evaluator(parse_results[0])
        right_evaluator = self._compose_evaluator(parse_results[2])

        def evaluator(root, variant_chunk_index):
            left_array = left_evaluator(root, variant_chunk_index)
            right_array = right_evaluator(root, variant_chunk_index)

            if parse_results[1] == "&":
                return np.logical_and(left_array, right_array)
            else:
                left_array = np.any(left_array, axis=tuple(range(1, left_array.ndim)))
                right_array = np.any(
                    right_array, axis=tuple(range(1, right_array.ndim))
                )
                return np.logical_and(left_array, right_array)

        return evaluator

    def _compose_or_evaluator(self, parse_results: pp.ParseResults) -> Callable:
        assert len(parse_results) == 3
        assert parse_results[1] in {"|", "||"}

        left_evaluator = self._compose_evaluator(parse_results[0])
        right_evaluator = self._compose_evaluator(parse_results[2])

        def evaluator(root, variant_chunk_index: int):
            left_array = left_evaluator(root, variant_chunk_index)
            right_array = right_evaluator(root, variant_chunk_index)

            if parse_results[1] == "|":
                return np.logical_or(left_array, right_array)
            else:
                left_array = np.any(left_array, axis=tuple(range(1, left_array.ndim)))
                right_array = np.any(
                    right_array, axis=tuple(range(1, right_array.ndim))
                )
                return np.logical_or(left_array, right_array)

        return evaluator

    def _compose_evaluator(self, parse_results: pp.ParseResults) -> Callable:
        results_name = parse_results.get_name()
        return self._composers[results_name](parse_results)



import numpy as np
import pytest

from vcztools import _vcztools

FIXED_FIELD_NAMES = ["chrom", "pos", "id", "qual", "ref", "alt", "filter"]


def example_fixed_data(num_variants, num_samples=0):
    chrom = np.array(["X"] * num_variants, dtype="S")
    pos = np.arange(num_variants, dtype=np.int32)
    id = np.array(["."] * num_variants, dtype="S").reshape((num_variants, 1))
    ref = np.array(["A"] * num_variants, dtype="S")
    alt = np.array(["T"] * num_variants, dtype="S").reshape((num_variants, 1))
    qual = np.arange(num_variants, dtype=np.float32)
    filter_ = np.ones(num_variants, dtype=bool).reshape((num_variants, 1))
    filter_id = np.array(["PASS"], dtype="S")
    return {
        "chrom": chrom,
        "pos": pos,
        "id": id,
        "qual": qual,
        "ref": ref,
        "alt": alt,
        "filter": filter_,
        "filter_ids": filter_id,
    }


def example_info_data(num_variants):
    d = {}
    for num_columns in range(1, 3):
        for dtype in ["i1", "i2", "i4", "f4", "?", "S1"]:
            name = f"I{dtype}_{num_columns}"
            data = np.arange(num_variants * num_columns).astype(dtype)
            d[name] = data.reshape((num_variants, num_columns))
    return d


def example_format_data(num_variants, num_samples):
    d = {}
    for num_columns in range(1, 3):
        for dtype in ["i1", "i2", "i4", "f4", "?", "S1"]:
            name = f"F{dtype}_{num_columns}"
            data = np.arange(num_variants * num_samples * num_columns).astype(dtype)
            d[name] = data.reshape((num_variants, num_samples, num_columns))
    return d


def example_gt_data(num_variants, num_samples, ploidy=2):
    return {
        "gt": np.arange(num_variants * num_samples * ploidy)
        .astype(np.int32)
        .reshape(num_variants, num_samples, ploidy),
        "gt_phased": (np.arange(num_variants * num_samples) % 2)
        .reshape(num_variants, num_samples)
        .astype(bool),
    }


def example_encoder(num_variants=1, num_samples=0, add_info=True):
    encoder = _vcztools.VcfEncoder(
        num_variants, num_samples, **example_fixed_data(num_variants, num_samples)
    )
    if add_info:
        for name, data in example_info_data(num_variants).items():
            encoder.add_info_field(name, data)
    if num_samples > 0:
        for name, data in example_format_data(num_variants, num_samples).items():
            encoder.add_format_field(name, data)

        gt_data = example_gt_data(num_variants, num_samples)
        encoder.add_gt_field(gt_data["gt"], gt_data["gt_phased"])
    # import sys
    # encoder.print_state(sys.stdout)
    return encoder


class TestPrintState:
    def test_nomimal_case(self, tmp_path):
        encoder = example_encoder()
        filename = tmp_path / "debug.txt"
        with open(filename, "w") as f:
            encoder.print_state(f)
        with open(filename) as f:
            s = f.read()
        assert "CHROM" in s

    def test_read_only_file(self, tmp_path):
        encoder = example_encoder()
        filename = tmp_path / "debug.txt"
        with open(filename, "w") as f:
            f.write("x")
        with open(filename) as f:
            with pytest.raises(OSError, match="22"):
                encoder.print_state(f)

    def test_no_arg(self):
        encoder = example_encoder()
        with pytest.raises(TypeError, match="function takes"):
            encoder.print_state()

    @pytest.mark.parametrize("fileobj", [None, "path"])
    def test_bad_file_arg(self, fileobj):
        encoder = example_encoder()
        with pytest.raises(TypeError):
            encoder.print_state(fileobj)

    @pytest.mark.parametrize("fileobj", [1024, 123])
    def test_bad_fileno(self, fileobj):
        encoder = example_encoder()
        with pytest.raises(OSError, match="9"):
            encoder.print_state(fileobj)


class TestEncode:
    @pytest.mark.parametrize("bad_arg", [None, {}, "0"])
    def test_bad_index_arg(self, bad_arg):
        encoder = example_encoder()
        with pytest.raises(TypeError, match="int"):
            encoder.encode(bad_arg, 1)

    @pytest.mark.parametrize("bad_arg", [None, {}, "0"])
    def test_bad_len_arg(self, bad_arg):
        encoder = example_encoder()
        with pytest.raises(TypeError, match="int"):
            encoder.encode(0, bad_arg)

    @pytest.mark.parametrize("bad_row", [2, 3, 100, -1, 2**32, 2**63])
    def test_bad_variant_arg(self, bad_row):
        encoder = example_encoder(2)
        with pytest.raises(ValueError, match="102"):
            encoder.encode(bad_row, 1)

    def test_buffer_malloc_fail(self):
        encoder = example_encoder(1)
        with pytest.raises(MemoryError):
            encoder.encode(0, 2**63)

    def test_small_example_overrun(self):
        encoder = example_encoder(1, 2)
        s = encoder.encode(0, 1024)
        minlen = len(s)
        for length in range(minlen):
            with pytest.raises(_vcztools.VczBufferTooSmall, match="-101"):
                encoder.encode(0, length)
        assert s == encoder.encode(0, minlen)

    @pytest.mark.parametrize("variant", [0, 999, 333, 501])
    def test_large_example_overrun(self, variant):
        encoder = example_encoder(1000, 100)
        s = encoder.encode(variant, 1024 * 1024)
        minlen = len(s)
        for length in range(minlen):
            with pytest.raises(_vcztools.VczBufferTooSmall, match="-101"):
                encoder.encode(variant, length)
        assert s == encoder.encode(variant, minlen)


class TestFixedFieldInputChecking:
    @pytest.mark.parametrize("name", FIXED_FIELD_NAMES)
    def test_field_incorrect_length(self, name):
        num_variants = 5
        data = example_fixed_data(num_variants)
        data[name] = data[name][1:]
        with pytest.raises(ValueError, match=f"Array {name} must have "):
            _vcztools.VcfEncoder(num_variants, 0, **data)

    @pytest.mark.parametrize("name", [*FIXED_FIELD_NAMES, "filter_ids"])
    def test_field_incorrect_dtype(self, name):
        num_variants = 5
        data = example_fixed_data(num_variants)
        data[name] = np.zeros(data[name].shape, dtype=np.int64)
        with pytest.raises(ValueError, match=f"Wrong dtype for {name}"):
            _vcztools.VcfEncoder(num_variants, 0, **data)

    @pytest.mark.parametrize("name", [*FIXED_FIELD_NAMES, "filter_ids"])
    def test_field_incorrect_dimension(self, name):
        num_variants = 2
        data = example_fixed_data(num_variants)
        data[name] = np.expand_dims(data[name], -1)
        with pytest.raises(ValueError, match=f"{name} has wrong dimension"):
            _vcztools.VcfEncoder(num_variants, 0, **data)

    @pytest.mark.parametrize("name", [*FIXED_FIELD_NAMES, "filter_ids"])
    def test_field_incorrect_type(self, name):
        num_variants = 5
        data = example_fixed_data(num_variants)
        data[name] = "A Python string"
        with pytest.raises(TypeError, match="must be numpy.ndarray"):
            _vcztools.VcfEncoder(num_variants, 0, **data)

    @pytest.mark.parametrize("name", ["id", "alt"])
    def test_zero_columns(self, name):
        data = example_fixed_data(1)
        arr = np.frombuffer(b"", dtype="S1").reshape(1, 0)
        data[name] = arr
        with pytest.raises(ValueError, match="-204"):
            _vcztools.VcfEncoder(1, 0, **data)

    def test_zero_filter_ids(self):
        data = example_fixed_data(1)
        data["filter_ids"] = np.array([], dtype="S")
        data["filter"] = np.array([], dtype=bool).reshape(1, 0)
        with pytest.raises(ValueError, match="-204"):
            _vcztools.VcfEncoder(1, 0, **data)

    def test_incorrect_num_filter_ids(self):
        data = example_fixed_data(1)
        data["filter_ids"] = np.array(["PASS", "1"], dtype="S")
        with pytest.raises(ValueError, match="filters dimension must be"):
            _vcztools.VcfEncoder(1, 0, **data)


class TestAddFields:
    def test_add_info_field_bad_num_args(self):
        encoder = example_encoder()
        with pytest.raises(TypeError):
            encoder.add_info_field()
        with pytest.raises(TypeError):
            encoder.add_info_field("name")

    def test_add_format_field_bad_num_args(self):
        encoder = example_encoder()
        with pytest.raises(TypeError):
            encoder.add_format_field()
        with pytest.raises(TypeError):
            encoder.add_format_field("name")

    def test_add_gt_field_bad_num_args(self):
        encoder = example_encoder()
        with pytest.raises(TypeError):
            encoder.add_gt_field()
        with pytest.raises(TypeError):
            encoder.add_gt_field(np.zeros((1), dtype=bool))

    @pytest.mark.parametrize("name", [None, 1234, {}])
    def test_add_info_field_bad_name_type(self, name):
        encoder = example_encoder()
        with pytest.raises(TypeError, match="must be str"):
            encoder.add_info_field(name, np.array([1]))

    @pytest.mark.parametrize("name", [None, 1234, {}])
    def test_add_format_field_bad_name_type(self, name):
        encoder = example_encoder()
        with pytest.raises(TypeError, match="must be str"):
            encoder.add_format_field(name, np.array([1]))

    @pytest.mark.parametrize("array", [None, 1234, {}])
    def test_add_info_field_bad_array_type(self, array):
        encoder = example_encoder()
        with pytest.raises(TypeError, match="ndarray"):
            encoder.add_info_field("name", array)

    @pytest.mark.parametrize("array", [None, 1234, {}])
    def test_add_format_field_bad_array_type(self, array):
        encoder = example_encoder()
        with pytest.raises(TypeError, match="ndarray"):
            encoder.add_format_field("name", array)

    @pytest.mark.parametrize("array", [None, 1234, {}])
    def test_add_gt_phased_field_bad_array_type(self, array):
        encoder = example_encoder(1, 1)
        with pytest.raises(TypeError, match="ndarray"):
            encoder.add_gt_field(np.zeros((1, 1, 1), dtype=np.int8), array)

    @pytest.mark.parametrize("array", [None, 1234, {}])
    def test_add_gt_field_bad_array_type(self, array):
        encoder = example_encoder(1, 1)
        with pytest.raises(TypeError, match="ndarray"):
            encoder.add_gt_field(array, np.zeros((1, 1), dtype=bool))

    def test_add_info_field_bad_array_num_variants(self):
        num_variants = 10
        encoder = example_encoder(num_variants)
        with pytest.raises(ValueError, match="number of variants"):
            encoder.add_info_field("name", np.zeros((3, 1), dtype=bool))

    def test_add_format_field_bad_array_num_variants(self):
        num_variants = 10
        encoder = example_encoder(num_variants)
        with pytest.raises(ValueError, match="number of variants"):
            encoder.add_format_field("name", np.zeros((3, 1, 1), dtype=bool))

    def test_add_gt_field_bad_gt_array_num_variants(self):
        num_variants = 10
        encoder = example_encoder(num_variants, 1)
        with pytest.raises(ValueError, match="number of variants"):
            encoder.add_gt_field(
                np.zeros((2, 1, 1), dtype=np.int8), np.zeros((10, 1), dtype=bool)
            )

    def test_add_gt_field_bad_gt_phased_array_num_variants(self):
        num_variants = 10
        encoder = example_encoder(num_variants, 1)
        with pytest.raises(ValueError, match="number of variants"):
            encoder.add_gt_field(
                np.zeros((10, 1, 1), dtype=np.int8), np.zeros((2, 1), dtype=bool)
            )

    def test_add_format_field_bad_array_num_samples(self):
        encoder = example_encoder(5, 2)
        with pytest.raises(ValueError, match="number of samples"):
            encoder.add_format_field("name", np.zeros((5, 1, 1), dtype=bool))

    def test_add_gt_field_bad_gt_array_num_samples(self):
        encoder = example_encoder(5, 2)
        with pytest.raises(ValueError, match="number of samples"):
            encoder.add_gt_field(
                np.zeros((5, 1, 1), dtype=np.int8), np.zeros((5, 2), dtype=bool)
            )

    def test_add_gt_field_bad_gt_phased_array_num_samples(self):
        encoder = example_encoder(5, 2)
        with pytest.raises(ValueError, match="number of samples"):
            encoder.add_gt_field(
                np.zeros((5, 2, 1), dtype=np.int8), np.zeros((5, 1), dtype=bool)
            )

    @pytest.mark.parametrize("shape", [(), (5,), (5, 1, 1)])
    def test_add_info_field_wrong_dimension(self, shape):
        encoder = example_encoder(5)
        with pytest.raises(ValueError, match="wrong dimension"):
            encoder.add_info_field("name", np.zeros(shape, dtype=bool))

    @pytest.mark.parametrize("shape", [(), (5,), (5, 1), (5, 1, 1, 1)])
    def test_add_format_field_wrong_dimension(self, shape):
        encoder = example_encoder(5)
        with pytest.raises(ValueError, match="wrong dimension"):
            encoder.add_format_field("name", np.zeros(shape, dtype=bool))

    @pytest.mark.parametrize("shape", [(), (5,), (5, 2), (5, 2, 1, 1)])
    def test_add_gt_field_gt_wrong_dimension(self, shape):
        encoder = example_encoder(5, 2)
        with pytest.raises(ValueError, match="wrong dimension"):
            encoder.add_gt_field(
                np.zeros(shape, dtype=np.int8), np.zeros((5, 2), dtype=bool)
            )

    @pytest.mark.parametrize("shape", [(), (5,), (5, 2, 1), (5, 2, 1, 1)])
    def test_add_gt_field_gt_phased_wrong_dimension(self, shape):
        encoder = example_encoder(5, 2)
        with pytest.raises(ValueError, match="wrong dimension"):
            encoder.add_gt_field(
                np.zeros((5, 2, 1), dtype=np.int8), np.zeros(shape, dtype=bool)
            )

    @pytest.mark.parametrize("dtype", ["m", "c16", "O", "u1"])
    def test_add_info_field_unsupported_dtype(self, dtype):
        encoder = example_encoder(1)
        with pytest.raises(ValueError, match="unsupported array dtype"):
            encoder.add_info_field("name", np.zeros((1, 1), dtype=dtype))

    @pytest.mark.parametrize("dtype", ["m", "c16", "O", "u1"])
    def test_add_format_field_unsupported_dtype(self, dtype):
        encoder = example_encoder(1, 1)
        with pytest.raises(ValueError, match="unsupported array dtype"):
            encoder.add_format_field("name", np.zeros((1, 1, 1), dtype=dtype))

    @pytest.mark.parametrize("dtype", ["m", "c16", "O", "u1"])
    def test_add_gt_field_gt_unsupported_dtype(self, dtype):
        encoder = example_encoder(1, 1)
        with pytest.raises(ValueError, match="unsupported array dtype"):
            encoder.add_gt_field(
                np.zeros((1, 1, 1), dtype=dtype), np.zeros((1, 1), dtype=bool)
            )

    @pytest.mark.parametrize("dtype", ["m", "c16", "O", "u1"])
    def test_add_gt_field_gt_phased_unsupported_dtype(self, dtype):
        encoder = example_encoder(1, 1)
        with pytest.raises(ValueError, match="Wrong dtype for gt_phased"):
            encoder.add_gt_field(
                np.zeros((1, 1, 1), dtype=np.int8), np.zeros((1, 1), dtype=dtype)
            )

    @pytest.mark.parametrize("dtype", ["i8", "f8"])
    def test_add_info_field_unsupported_width(self, dtype):
        encoder = example_encoder(1)
        with pytest.raises(ValueError, match="-203"):
            encoder.add_info_field("name", np.zeros((1, 1), dtype=dtype))

    @pytest.mark.parametrize("dtype", ["i8", "f8"])
    def test_add_format_field_unsupported_width(self, dtype):
        encoder = example_encoder(1, 1)
        with pytest.raises(ValueError, match="-203"):
            encoder.add_format_field("name", np.zeros((1, 1, 1), dtype=dtype))

    def test_add_gt_field_unsupported_width(self):
        encoder = example_encoder(1, 1)
        with pytest.raises(ValueError, match="-203"):
            encoder.add_gt_field(
                np.zeros((1, 1, 1), dtype=np.int64), np.zeros((1, 1), dtype=bool)
            )

    def test_add_gt_field_zero_ploidy(self):
        encoder = example_encoder(1, 1)
        with pytest.raises(ValueError, match="-204"):
            encoder.add_gt_field(
                np.zeros((1, 1, 0), dtype=np.int64), np.zeros((1, 1), dtype=bool)
            )


class TestArrays:
    def test_stored_data_equal(self):
        num_variants = 20
        num_samples = 10
        fixed_data = example_fixed_data(num_variants, num_samples)
        gt_data = example_gt_data(num_variants, num_samples)
        encoder = _vcztools.VcfEncoder(num_variants, num_samples, **fixed_data)
        encoder.add_gt_field(gt_data["gt"], gt_data["gt_phased"])
        info_data = example_info_data(num_variants)
        format_data = example_format_data(num_variants, num_samples)
        for name, data in info_data.items():
            encoder.add_info_field(name, data)
        for name, data in format_data.items():
            encoder.add_format_field(name, data)
        all_data = {**fixed_data}
        for name, array in info_data.items():
            all_data[f"INFO/{name}"] = array
        for name, array in {**format_data, **gt_data}.items():
            all_data[f"FORMAT/{name}"] = array
        encoder_arrays = encoder.arrays
        assert set(encoder.arrays.keys()) == set(all_data.keys())
        for name in all_data.keys():
            # Strong identity assertion here for now, but this may change
            assert all_data[name] is encoder_arrays[name]

    def test_array_bad_alignment(self):
        fixed_data = example_fixed_data(1)
        a = fixed_data["pos"]
        # Value is not aligned
        a = np.frombuffer(b"\0\0\0\0\0", dtype=np.int32, offset=1)
        fixed_data["pos"] = a
        with pytest.raises(ValueError, match="NPY_ARRAY_IN_ARRAY"):
            _vcztools.VcfEncoder(1, 0, **fixed_data)


class TestUninitialised:
    @pytest.mark.parametrize(
        "name",
        ["add_info_field", "add_gt_field", "add_format_field", "print_state", "encode"],
    )
    def test_methods(self, name):
        cls = _vcztools.VcfEncoder
        encoder = cls.__new__(cls)
        method = getattr(encoder, name)
        with pytest.raises(SystemError, match="not initialised"):
            method()

    @pytest.mark.parametrize("name", ["arrays"])
    def test_attrs(self, name):
        cls = _vcztools.VcfEncoder
        encoder = cls.__new__(cls)
        with pytest.raises(SystemError, match="not initialised"):
            getattr(encoder, name)



import numpy
from setuptools import Extension, setup

_vcztools_module = Extension(
    "vcztools._vcztools",
    sources=["vcztools/_vcztoolsmodule.c", "lib/vcf_encoder.c"],
    extra_compile_args=["-std=c99"],
    include_dirs=["lib", numpy.get_include()],
)

setup(
    name="vcztools",
    ext_modules=[_vcztools_module],
)
